namespace Commonmark {

    /**
     * The node renderer that renders all the core nodes (comes last in the
     * order of node renderers).
     */
    public class CoreHtmlNodeRenderer : AbstractVisitor, NodeRenderer {

        protected HtmlNodeRendererContext context;
        private HtmlWriter html;

        public CoreHtmlNodeRenderer(HtmlNodeRendererContext context) {
            this.context = context;
            this.html = context.get_writer();
        }

        public Gee.Set<GLib.Type> get_node_types() {
            var nt = new Gee.HashSet<GLib.Type>();
            nt.add_all_array(
                new GLib.Type[] {
                    typeof(Document),
                    typeof(Heading),
                    typeof(Paragraph),
                    typeof(BlockQuote),
                    typeof(BulletList),
                    typeof(FencedCodeBlock),
                    typeof(HtmlBlock),
                    typeof(ThematicBreak),
                    typeof(IndentedCodeBlock),
                    typeof(Link),
                    typeof(ListItem),
                    typeof(OrderedList),
                    typeof(Image),
                    typeof(Emphasis),
                    typeof(StrongEmphasis),
                    typeof(Text),
                    typeof(Code),
                    typeof(HtmlInline),
                    typeof(SoftLineBreak),
                    typeof(HardLineBreak) });
            return nt;
        }

        public void render(Node node) {
            node.accept(this);
        }

        public override void visit(Node node) {
            if (node is BlockQuote)
                visit_block_quote(node as BlockQuote);
            else if (node is BulletList)
                visit_bullet_list(node as BulletList);
            else if (node is Code)
                visit_code(node as Code);
            else if (node is Document)
                visit_children(node as Document);
            else if (node is Emphasis)
                visit_emphasis(node as Emphasis);
            else if (node is FencedCodeBlock)
                visit_fenced_code_block(node as FencedCodeBlock);
            else if (node is HardLineBreak)
                visit_hard_line_break(node as HardLineBreak);
            else if (node is Heading)
                visit_heading(node as Heading);
            else if (node is HtmlBlock)
                visit_html_block(node as HtmlBlock);
            else if (node is HtmlInline)
                visit_html_inline(node as HtmlInline);
            else if (node is Image)
                visit_image(node as Image);
            else if (node is IndentedCodeBlock)
                visit_indented_code_block(node as IndentedCodeBlock);
            else if (node is Link)
                visit_link(node as Link);
            else if (node is ListItem)
                visit_list_item(node as ListItem);
            else if (node is OrderedList)
                visit_ordered_list(node as OrderedList);
            else if (node is Paragraph)
                visit_paragraph(node as Paragraph);
            else if (node is SoftLineBreak)
                visit_soft_line_break(node as SoftLineBreak);
            else if (node is StrongEmphasis)
                visit_strong_emphasis(node as StrongEmphasis);
            else if (node is Text)
                visit_text(node as Text);
            else if (node is ThematicBreak)
                visit_thematic_break(node as ThematicBreak);
            else
                GLib.error("Invalid node type");
        }

        private void visit_block_quote(BlockQuote block_quote) {
            html.line();
            html.tag_attrs("blockquote", get_attrs(block_quote, "blockquote"));
            html.line();
            visit_children(block_quote);
            html.line();
            html.tag("/blockquote");
            html.line();
        }

        private void visit_bullet_list(BulletList bullet_list) {
            render_list_block(bullet_list, "ul",
                              get_attrs(bullet_list, "ul"));
        }

        private void visit_code(Code code) {
            html.tag_attrs("code", get_attrs(code, "code"));
            html.text(code.literal);
            html.tag("/code");
        }

        private void visit_emphasis(Emphasis emphasis) {
            html.tag_attrs("em", get_attrs(emphasis, "em"));
            visit_children(emphasis);
            html.tag("/em");
        }

        private void visit_fenced_code_block(
            FencedCodeBlock fenced_code_block) {
            string literal = fenced_code_block.literal;
            var attributes = new LinkedMap<string, string>();
            string info = fenced_code_block.info;
            if (info != null && info.length > 0) {
                int space = info.index_of(" ");
                string language;
                if (space == -1) {
                    language = info;
                } else {
                    language = info.substring(0, space);
                }
                attributes["class"] = "language-" + language;
            }
            render_code_block(literal, fenced_code_block, attributes);
        }

        private void visit_hard_line_break(
            HardLineBreak hard_line_break) {
            html.tag_full("br", get_attrs(hard_line_break, "br"), true);
            html.line();
        }

        private void visit_heading(Heading heading) {
            string htag = "h%d".printf(heading.level);
            html.line();
            html.tag_attrs(htag, get_attrs(heading, htag));
            visit_children(heading);
            html.tag("/" + htag);
            html.line();
        }

        private void visit_html_block(HtmlBlock html_block) {
            html.line();
            if (context.should_escape_html()) {
                html.tag_attrs("p", get_attrs(html_block, "p"));
                html.text(html_block.literal);
                html.tag("/p");
            } else {
                html.raw(html_block.literal);
            }
            html.line();
        }

        private void visit_html_inline(HtmlInline html_inline) {
            if (context.should_escape_html()) {
                html.text(html_inline.literal);
            } else {
                html.raw(html_inline.literal);
            }
        }

        private void visit_image(Image image) {
            string url = image.destination;

            var alt_text_visitor = new AltTextVisitor();
            image.accept(alt_text_visitor);
            string alt_text = alt_text_visitor.get_alt_text();

            var attrs = new LinkedMap<string, string>();
            if (context.should_sanitize_urls()) {
                url = context.url_sanitizer().sanitize_image_url(url);
            }

            attrs["src"] = context.encode_url(url);
            attrs["alt"] = alt_text;
            if (image.title != null) {
                attrs["title"] = image.title;
            }

            html.tag_full("img", get_attrs_with_default(image, "img", attrs),
                          true);
        }

        private void visit_indented_code_block(
            IndentedCodeBlock indented_code_block) {
            render_code_block(indented_code_block.literal,
                              indented_code_block,
                              new LinkedMap<string, string>());
        }

        private void visit_link(Link link) {
            var attrs = new LinkedMap<string, string>();
            string url = link.destination;

            if (context.should_sanitize_urls()) {
                url = context.url_sanitizer().sanitize_link_url(url);
                attrs["rel"] = "nofollow";
            }

            url = context.encode_url(url);
            attrs["href"] = url;
            if (link.title != null) {
                attrs["title"] = link.title;
            }
            html.tag_attrs("a", get_attrs_with_default(link, "a", attrs));
            visit_children(link);
            html.tag("/a");
        }

        private void visit_list_item(ListItem list_item) {
            html.tag_attrs("li", get_attrs(list_item, "li"));
            visit_children(list_item);
            html.tag("/li");
            html.line();
        }

        private void visit_ordered_list(OrderedList ordered_list) {
            int start = ordered_list.start_number;
            var attrs = new LinkedMap<string, string>();
            if (start != 1) {
                attrs["start"] = "%d".printf(start);
            }
            render_list_block(ordered_list, "ol",
                              get_attrs_with_default(ordered_list, "ol",
                                                     attrs));
        }

        private void visit_paragraph(Paragraph paragraph) {
            bool in_tight_list = is_in_tight_list(paragraph);
            if (!in_tight_list) {
                html.line();
                html.tag_attrs("p", get_attrs(paragraph, "p"));
            }
            visit_children(paragraph);
            if (!in_tight_list) {
                html.tag("/p");
                html.line();
            }
        }

        private void visit_soft_line_break(SoftLineBreak softLineBreak) {
            html.raw(context.get_softbreak());
        }

        private void visit_strong_emphasis(
            StrongEmphasis strong_emphasis) {
            html.tag_attrs("strong", get_attrs(strong_emphasis, "strong"));
            visit_children(strong_emphasis);
            html.tag("/strong");
        }

        private void visit_text(Text text) {
            html.text(text.literal);
        }

        private void visit_thematic_break(ThematicBreak thematicBreak) {
            html.line();
            html.tag_full("hr", get_attrs(thematicBreak, "hr"), true);
            html.line();
        }

        protected override void visit_children(Node parent) {
            var node = parent.first_child;
            while (node != null) {
                context.render(node);
                node = node.next;
            }
        }

        private void render_code_block(string literal,
                                       Node node,
                                       Gee.Map<string, string> attributes) {
            html.line();
            html.tag_attrs("pre", get_attrs(node, "pre"));
            html.tag_attrs("code", get_attrs_with_default(node, "code",
                                                          attributes));
            html.text(literal);
            html.tag("/code");
            html.tag("/pre");
            html.line();
        }

        private void render_list_block(ListBlock list_block,
                                       string tag_name,
                                       Gee.Map<string, string> attributes) {
            html.line();
            html.tag_attrs(tag_name, attributes);
            html.line();
            visit_children(list_block);
            html.line();
            html.tag("/" + tag_name);
            html.line();
        }

        private bool is_in_tight_list(Paragraph paragraph) {
            var parent = paragraph.parent;
            if (parent != null) {
                var gramps = parent.parent;
                if (gramps is ListBlock) {
                    ListBlock list = gramps as ListBlock;
                    return list.tight;
                }
            }
            return false;
        }

        private Gee.Map<string, string> get_attrs(Node node,
                                                  string tag_name) {
            return get_attrs_with_default(node, tag_name,
                                          new LinkedMap<string, string>());
        }

        private Gee.Map<string, string>
            get_attrs_with_default(Node node, string tag_name,
                                   Gee.Map<string, string> default_attributes) {
            return context.extend_attributes(node, tag_name, default_attributes);
        }

        private class AltTextVisitor : AbstractVisitor {

            private StringBuilder sb;

            construct {
                sb = new StringBuilder();
            }

            public string get_alt_text() {
                return sb.str;
            }

            public override void visit(Node node) {
                if (node is Text)
                    sb.append(((Text)node).literal);
                else if (node is SoftLineBreak || node is HardLineBreak)
                    sb.append_c('\n');
                else {
                    base.visit(node);
                }
            }
        }
    }
}
