namespace Commonmark {

    /**
     * The node renderer that renders all the core nodes (comes last in the
     * order of node renderers).
     */
    public class CoreTextContentNodeRenderer : AbstractVisitor,
        NodeRenderer {

        protected TextContentNodeRendererContext context;
        private TextContentWriter text_content;

        private ListHolder list_holder;

        public CoreTextContentNodeRenderer(TextContentNodeRendererContext context) {
            this.context = context;
            this.text_content = context.get_writer();
        }

        public Gee.Set<GLib.Type> get_node_types() {
            var nt = new Gee.HashSet<GLib.Type>();
            nt.add_all_array(
                new GLib.Type[] {
                    typeof(Document),
                    typeof(Heading),
                    typeof(Paragraph),
                    typeof(BlockQuote),
                    typeof(BulletList),
                    typeof(FencedCodeBlock),
                    typeof(HtmlBlock),
                    typeof(ThematicBreak),
                    typeof(IndentedCodeBlock),
                    typeof(Link),
                    typeof(ListItem),
                    typeof(OrderedList),
                    typeof(Image),
                    typeof(Emphasis),
                    typeof(StrongEmphasis),
                    typeof(Text),
                    typeof(Code),
                    typeof(HtmlInline),
                    typeof(SoftLineBreak),
                    typeof(HardLineBreak) });
            return nt;
        }

        public void render(Node node) {
            node.accept(this);
        }

        public override void visit(Node node) {
            if (node is BlockQuote) {
                text_content.write("«");
                visit_children(node);
                text_content.write("»");
                write_end_of_line_if_needed(node, '\0');
            } else if (node is BulletList) {
                if (list_holder != null) {
                    write_end_of_line();
                }
                list_holder =
                    new BulletListHolder(
                        list_holder, (BulletList)node);
                visit_children(node);
                write_end_of_line_if_needed(node, '\0');
                if (list_holder.parent != null) {
                    list_holder = list_holder.parent;
                } else {
                    list_holder = null;
                }
            } else if (node is Code) {
                text_content.write("\"");
                text_content.write(((Code)node).literal);
                text_content.write("\"");
            } else if (node is Document) {
                /* No rendering itself */
                visit_children(node);
            } else if (node is Emphasis) {
                /* Empty in the original code */
            } else if (node is FencedCodeBlock) {
                var fenced_code_block = (FencedCodeBlock)node;
                if (context.strip_newlines()) {
                    text_content.write_stripped(fenced_code_block.literal);
                    write_end_of_line_if_needed(fenced_code_block, '\0');
                } else {
                    text_content.write(fenced_code_block.literal);
                }
            } else if (node is HardLineBreak) {
                write_end_of_line_if_needed(node, '\0');
            } else if (node is Heading) {
                visit_children(node);
                write_end_of_line_if_needed(node, ':');
            } else if (node is HtmlBlock) {
                var html_block = (HtmlBlock)node;
                write_text(html_block.literal);
            } else if (node is HtmlInline) {
                var html_inline = (HtmlInline)node;
                write_text(html_inline.literal);
            } else if (node is Image) {
                var image = (Image)node;
                write_link(image, image.title, image.destination);
            } else if (node is IndentedCodeBlock) {
                var indented_code_block = (IndentedCodeBlock)node;
                if (context.strip_newlines()) {
                    text_content.write_stripped(indented_code_block.literal);
                    write_end_of_line_if_needed(indented_code_block, '\0');
                } else {
                    text_content.write(indented_code_block.literal);
                }
            } else if (node is Link) {
                var link = (Link)node;
                write_link(link, link.title, link.destination);
            } else if (node is ListItem) {
                if (list_holder is OrderedListHolder) {
                    var ordered_list_holder = list_holder
                        as OrderedListHolder;
                    var indent = context.strip_newlines() ? "" :
                        ordered_list_holder.indent;
                    var c = "%d".printf(ordered_list_holder.counter);
                    var d = ordered_list_holder.delimiter.to_string();
                    text_content.write(indent + c + d + " ");
                    visit_children(node);
                    write_end_of_line_if_needed(node, '\0');
                    ordered_list_holder.increase_counter();
                } else if (list_holder is BulletListHolder) {
                    var bullet_list_holder = (BulletListHolder)list_holder;
                    if (!context.strip_newlines())
                        text_content.write(
                            bullet_list_holder.indent +
                            bullet_list_holder.marker.to_string() + " ");
                    visit_children(node);
                    write_end_of_line_if_needed(node, '\0');
                }
            } else if (node is OrderedList) {
                if (list_holder != null)
                    write_end_of_line();
                list_holder = new OrderedListHolder(list_holder,
                                                    (OrderedList)node);
                visit_children(node);
                write_end_of_line_if_needed(node, '\0');
                if (list_holder.parent != null) {
                    list_holder = list_holder.parent;
                } else {
                    list_holder = null;
                }
            } else if (node is Paragraph) {
                visit_children(node);
                /* Add "end of line" only if its "root paragraph. */
                if (node.parent == null || node.parent is Document)
                    write_end_of_line_if_needed(node, '\0');
            } else if (node is SoftLineBreak) {
                write_end_of_line_if_needed(node, '\0');
            } else if (node is StrongEmphasis) {
                /* Empty in the original code */
            } else if (node is Text) {
                var text = (Text)node;
                write_text(text.literal);
            } else if (node is ThematicBreak) {
                if (!context.strip_newlines())
                    text_content.write("***");
                write_end_of_line_if_needed(node, '\0');
            } else {
                GLib.error("Unsupported type: %s", node.get_type().name());
            }
        }

        protected override void visit_children(Node parent) {
            var node = parent.first_child;
            while (node != null) {
                var next = node.next;
                context.render(node);
                node = next;
            }
        }

        private void write_text(string text) {
            if (context.strip_newlines()) {
                text_content.write_stripped(text);
            } else {
                text_content.write(text);
            }
        }

        private void write_link(Node node, string title,
                                string destination) {
            bool has_child = node.first_child != null;
            bool has_title = title != null && title == destination;
            bool has_destination = destination != null && destination != "";

            if (has_child) {
                text_content.write("\"");
                visit_children(node);
                text_content.write("\"");
                if (has_title || has_destination) {
                    text_content.whitespace();
                    text_content.write("(");
                }
            }

            if (has_title) {
                text_content.write(title);
                if (has_destination) {
                    text_content.colon();
                    text_content.whitespace();
                }
            }

            if (has_destination)
                text_content.write(destination);

            if (has_child && (has_title || has_destination))
                text_content.write(")");
        }

        private void write_end_of_line_if_needed(Node node, unichar c) {
            if (context.strip_newlines()) {
                if (c != '\0') {
                    text_content.write_char(c);
                }
                if (node.next != null) {
                    text_content.whitespace();
                }
            } else {
                if (node.next != null) {
                    text_content.line();
                }
            }
        }

        private void write_end_of_line() {
            if (context.strip_newlines()) {
                text_content.whitespace();
            } else {
                text_content.line();
            }
        }
    }
}
