namespace Commonmark {

    public class DocumentParser : GLib.Object, ParserState {

        private static Gee.Set<GLib.Type> CORE_FACTORY_TYPES;

        private static Gee.Map<GLib.Type, BlockParserFactory>
            NODES_TO_CORE_FACTORIES;

        private SourceLine line;

        /**
         * Line index (0-based)
         */
        private int line_index;

        /**
         * current index (offset) in input line (0-based)
         */
        private int index;

        /**
         * current column of input line (tab causes column to go to next 4-space
         * tab stop) (0-based)
         */
        private int column;

        /**
         * if the current column is within a tab character (partially consumed
         * tab)
         */
        private bool column_is_in_tab;

        private int next_non_space;
        private int next_non_space_column;
        private int indent;
        private bool blank;

        private Gee.List<BlockParserFactory>
            block_parser_factories;
        private InlineParserFactory inline_parser_factory;
        private Gee.List<DelimiterProcessor>
            delimiter_processors;
        private IncludeSourceSpans include_source_spans;
        private DocumentBlockParser document_block_parser;
        private LinkReferenceDefinitions definitions;

        private Gee.List<OpenBlockParser> open_block_parsers;
        private Gee.List<BlockParser> all_block_parsers;

        public DocumentParser.from() {}

        public DocumentParser(
            Gee.List<BlockParserFactory> block_parser_factories,
            InlineParserFactory inline_parser_factory,
            Gee.List<DelimiterProcessor> delimiter_processors,
            IncludeSourceSpans include_source_spans) {
            this.block_parser_factories = block_parser_factories;
            this.inline_parser_factory = inline_parser_factory;
            this.delimiter_processors = delimiter_processors;
            this.include_source_spans = include_source_spans;

            line_index = -1;

            definitions = new LinkReferenceDefinitions();
            open_block_parsers = new Gee.ArrayList<OpenBlockParser>();
            all_block_parsers = new Gee.ArrayList<BlockParser>();

            this.document_block_parser = new DocumentBlockParser();
            var obp = new OpenBlockParser(document_block_parser, 0);
            activate_block_parser(obp);
        }

        public static Gee.Set<GLib.Type> get_default_block_parser_types() {
            return CORE_FACTORY_TYPES;
        }

        public static Gee.List<BlockParserFactory>
            calculate_block_parser_factories(
                Gee.List<BlockParserFactory>
                custom_block_parser_factories,
                Gee.Set<GLib.Type> enabled_block_types) {
            var list = new Gee.ArrayList<BlockParserFactory>();
            /* By having the custom factories come first, extensions are able to
               change behavior of core syntax. */
            list.add_all(custom_block_parser_factories);
            foreach (var block_type in enabled_block_types) {
                list.add(NODES_TO_CORE_FACTORIES[block_type]);
            }
            return list;
        }

        public static void check_enabled_block_types(
            Gee.Set<GLib.Type> enabled_block_types) {
            foreach (var enabled_block_type in enabled_block_types) {
                if (!NODES_TO_CORE_FACTORIES.has_key(enabled_block_type)) {
                    var vs = "";
                    foreach (var v in NODES_TO_CORE_FACTORIES.keys)
                        vs += "%s, ".printf(v.name());
                    GLib.error(
                        "Can't enable block type %s, possible options are: %s",
                        enabled_block_type.name(), vs);
                }
            }
        }

        /**
         * The main parsing function. Returns a parsed document AST.
         */
        public Document parse(string input) {
            int line_start = 0;
            int line_break;
            while ((line_break =
                    Parsing.find_line_break(input, line_start)) != -1) {
                var line = input.slice(line_start, line_break);
                parse_line(line);
                if (line_break + 1 < input.length &&
                    input.get(line_break) == '\r' &&
                    input.get(line_break + 1) == '\n') {
                    line_start = line_break + 2;
                } else {
                    line_start = line_break + 1;
                }
            }
            if (input.length > 0 &&
                (line_start == 0 || line_start < input.length)) {
                var line = input.slice(line_start, input.length);
                parse_line(line);
            }

            return finalize_and_process();
        }

        public Document parse_input(GLib.InputStream input)
            throws GLib.Error {

            var dis = new GLib.DataInputStream(input);

            string line;
            while ((line = dis.read_line(null)) != null)
                parse_line(line);

            return finalize_and_process();
        }

        public SourceLine get_line() {
            return line;
        }

        public int get_index() {
            return index;
        }

        public int get_next_non_space_index() {
            return next_non_space;
        }

        public int get_column() {
            return column;
        }

        public int get_indent() {
            return indent;
        }

        public bool is_blank() {
            return blank;
        }

        public BlockParser get_active_block_parser() {
            return open_block_parsers[open_block_parsers.size - 1].block_parser;
        }

        /**
         * Analyze a line of text and update the document appropriately. We
         * parse markdown text by calling this on each line of input, then
         * finalizing the document.
         */
        private void parse_line(string ln) {
            set_line(ln);

            /* For each containing block, try to parse the associated line
               start. The document will always match, so we can skip the first
               block parser and start at 1 matches */
            int matches = 1;
            for (int i = 1; i < open_block_parsers.size; i++) {
                var open_block_parser = open_block_parsers[i];
                var block_parser = open_block_parser.block_parser;
                find_next_non_space();

                var result = block_parser.try_continue(this);
                if (result is BlockContinueImpl) {
                    var block_continue = (BlockContinueImpl)result;
                    open_block_parser.source_index = get_index();
                    if (block_continue.is_finalize()) {
                        add_source_spans();
                        close_block_parsers(open_block_parsers.size - i);
                        return;
                    } else {
                        if (block_continue.get_new_index() != -1) {
                            set_new_index(block_continue.get_new_index());
                        } else if (block_continue.get_new_column() != -1) {
                            set_new_column(block_continue.get_new_column());
                        }
                        matches++;
                    }
                } else {
                    break;
                }
            }

            int unmatched_blocks = open_block_parsers.size - matches;
            var block_parser = open_block_parsers[matches - 1].block_parser;
            bool started_new_block = false;

            int last_index = index;

            /* Unless last matched container is a code block, try new container
               starts, adding children to the last matched container: */
            bool try_block_starts = (block_parser.get_block() is Paragraph ||
                                     block_parser.is_container());
            while (try_block_starts) {
                last_index = index;
                find_next_non_space();

                /* this is a little performance optimization: */
                if (is_blank() ||
                    (indent < Parsing.CODE_BLOCK_INDENT &&
                     Parsing.is_letter(this.line.content, next_non_space))) {
                    set_new_index(next_non_space);
                    break;
                }

                var block_start = find_block_start(block_parser);
                if (block_start == null) {
                    set_new_index(next_non_space);
                    break;
                }

                started_new_block = true;
                int source_index = get_index();

                /* We're starting a new block. If we have any previous blocks
                   that need to be closed, we need to do it now. */
                if (unmatched_blocks > 0) {
                    close_block_parsers(unmatched_blocks);
                    unmatched_blocks = 0;
                }

                if (block_start.get_new_index() != -1) {
                    set_new_index(block_start.get_new_index());
                } else if (block_start.get_new_column() != -1) {
                    set_new_column(block_start.get_new_column());
                }

                Gee.List<SourceSpan> replaced_source_spans = null;
                if (block_start.is_replace_active_block_parser()) {
                    var replaced_block =
                        prepare_active_block_parser_for_replacement();
                    replaced_source_spans = replaced_block.source_spans;
                }

                foreach (var new_block_parser in block_start.get_block_parsers()) {
                    add_child(new OpenBlockParser(new_block_parser, source_index));
                    if (replaced_source_spans != null) {
                        new_block_parser.get_block().source_spans =
                            replaced_source_spans;
                    }
                    block_parser = new_block_parser;
                    try_block_starts = new_block_parser.is_container();
                }
            }

            /* What remains at the offset is a text line. Add the text to the
               appropriate block. */

            /* First check for a lazy paragraph continuation: */
            if (!started_new_block && !is_blank() &&
                get_active_block_parser().can_have_lazy_continuation_lines()) {
                open_block_parsers[open_block_parsers.size - 1].source_index =
                    last_index;
                /* lazy paragraph continuation */
                add_line();

            } else {

                /* finalize any blocks not matched */
                if (unmatched_blocks > 0) {
                    close_block_parsers(unmatched_blocks);
                }

                if (!block_parser.is_container()) {
                    add_line();
                } else if (!is_blank()) {
                    /* create paragraph container for line */
                    var paragraph_parser = new ParagraphParser();
                    add_child(new OpenBlockParser(paragraph_parser,
                                                  last_index));
                    add_line();
                } else {
                    /* This can happen for a list item like this:
                       ```
                       *
                       list item
                       ```
                       The first line does not start a paragraph yet, but we
                       still want to record source positions. */
                    add_source_spans();
                }
            }
        }

        private void set_line(string ln) {
            line_index++;
            index = 0;
            column = 0;
            column_is_in_tab = false;

            string line_content = Parsing.prepare_line(ln);
            SourceSpan source_span = null;
            if (include_source_spans != IncludeSourceSpans.NONE)
                source_span =
                    SourceSpan.from(line_index, 0, line_content.length);
            this.line = SourceLine.from(line_content, source_span);
        }

        private void find_next_non_space() {
            int i = index;
            int cols = column;

            blank = true;
            int length = line.content.length;
            while (i < length) {
                char c = line.content.get(i);
                switch (c) {
                case ' ':
                    i++;
                    cols++;
                    continue;
                case '\t':
                    i++;
                    cols += (4 - (cols % 4));
                    continue;
                }
                blank = false;
                break;
            }

            next_non_space = i;
            next_non_space_column = cols;
            indent = next_non_space_column - column;
        }

        private void set_new_index(int new_index) {
            if (new_index >= next_non_space) {
                /* We can start from here, no need to calculate tab stops
                   again */
                index = next_non_space;
                column = next_non_space_column;
            }
            int length = line.content.length;
            while (index < new_index && index != length)
                advance();
            /* If we're going to an index as opposed to a column, we're never
               within a tab */
            column_is_in_tab = false;
        }

        private void set_new_column(int new_column) {
            if (new_column >= next_non_space_column) {
                /* We can start from here, no need to calculate tab stops
                   again */
                index = next_non_space;
                column = next_non_space_column;
            }
            int length = line.content.length;
            while (column < new_column && index != length) {
                advance();
            }
            if (column > new_column) {
                /* Last character was a tab and we overshot our target */
                index--;
                column = new_column;
                column_is_in_tab = true;
            } else {
                column_is_in_tab = false;
            }
        }

        private void advance() {
            var c = line.content.get(index);
            index++;
            if (c == '\t') {
                column +=
                    Parsing.columns_to_next_tab_stop(column);
            } else {
                column++;
            }
        }

        /**
         * Add line content to the active block parser. We assume it can accept
         * lines -- that check should be done before calling this.
         */
        private void add_line() {
            string content;
            if (column_is_in_tab) {
                /* Our column is in a partially consumed tab. Expand the
                   remaining columns (to the next tab stop) to spaces. */
                int after_tab = index + 1;
                var rest = line.content.slice(after_tab, line.content.length);
                int spaces =
                    Parsing.columns_to_next_tab_stop(column);
                var sb = new StringBuilder();
                sb.append("%d".printf(spaces + rest.length));
                for (int i = 0; i < spaces; i++)
                    sb.append_c(' ');
                sb.append(rest);
                content = sb.str;
            } else if (index == 0) {
                content = line.content;
            } else {
                content = line.content.slice(index, line.content.length);
            }
            SourceSpan source_span = null;
            if (include_source_spans ==
                IncludeSourceSpans.BLOCKS_AND_INLINES) {
                /* Note that if we're in a partially-consumed tab, the length
                   here corresponds to the content but not to the actual source
                   length. That sounds like a problem, but I haven't found a
                   test case where it matters (yet). */
                source_span =
                    SourceSpan.from(line_index, index, content.length);
            }
            get_active_block_parser().add_line(
                SourceLine.from(content, source_span));
            add_source_spans();
        }

        private void add_source_spans() {
            if (include_source_spans != IncludeSourceSpans.NONE) {
                /* Don't add source spans for Document itself (it would get the
                   whole source text) */
                for (int i = 1; i < open_block_parsers.size; i++) {
                    var open_block_parser = open_block_parsers[i];
                    int block_index = open_block_parser.source_index;
                    int length = line.content.length - block_index;
                    if (length != 0)
                        open_block_parser.block_parser.add_source_span(
                            SourceSpan.from(line_index, block_index,
                                                 length));
                }
            }
        }

        private BlockStartImpl?
            find_block_start(BlockParser block_parser) {
            var matched_block_parser = new MatchedBlockParserImpl(block_parser);
            foreach (var block_parser_factory in block_parser_factories) {
                var result = block_parser_factory
                    .try_start(this, matched_block_parser);
                if (result is BlockStartImpl) {
                    return (BlockStartImpl)result;
                }
            }
            return null;
        }

        /**
         * Finalize a block. Close it and do any necessary postprocessing,
         * e.g. setting the content of blocks and collecting link reference
         * definitions from paragraphs.
         */
        private void do_finalize(BlockParser block_parser) {
            if (block_parser is ParagraphParser) {
                add_definitions_from((ParagraphParser)block_parser);
            }

            block_parser.close_block();
        }

        private void add_definitions_from(ParagraphParser paragraph_parser) {
            foreach (var definition in paragraph_parser.get_definitions()) {
                /* Add nodes into document before paragraph. */
                paragraph_parser.get_block().insert_before(definition);
                definitions.add(definition);
            }
        }

        /**
         * Walk through a block & children recursively, parsing string content
         * into inline content where appropriate.
         */
        private void process_inlines() {
            var context = new InlineParserContextImpl(delimiter_processors, definitions);
            var inline_parser = inline_parser_factory.create(context);

            foreach (var block_parser in all_block_parsers) {
                block_parser.parse_inlines(inline_parser);
            }
        }

        /**
         * Add block of type tag as a child of the tip. If the tip can't accept
         * children, close and finalize it and try its parent, and so on until
         * we find a block that can accept children.
         */
        private void add_child(OpenBlockParser open_block_parser) {
            while (!get_active_block_parser().can_contain(
                       open_block_parser.block_parser.get_block())) {
                close_block_parsers(1);
            }

            get_active_block_parser().get_block().append_child(
                open_block_parser.block_parser.get_block());
            activate_block_parser(open_block_parser);
        }

         private void activate_block_parser(OpenBlockParser open_block_parser) {
             open_block_parsers.add(open_block_parser);
         }

        private OpenBlockParser deactivate_block_parser() {
            return open_block_parsers.remove_at(open_block_parsers.size - 1);
        }

        private Block prepare_active_block_parser_for_replacement() {
            /* Note that we don't want to parse inlines, as it's getting
               replaced. */
            var old = deactivate_block_parser().block_parser;

            if (old is ParagraphParser) {
                var paragraph_parser = old as ParagraphParser;
                /* Collect any link reference definitions. Note that replacing
                   the active block parser is done after a block parser got the
                   current paragraph content using
                   MatchedBlockParser#getContentString. In case the paragraph
                   started with link reference definitions, we parse and strip
                   them before the block parser gets the content. We want to
                   keep them.  If no replacement happens, we collect the
                   definitions as part of finalizing paragraph blocks. */
                add_definitions_from(paragraph_parser);
            }

            /* Do this so that source positions are calculated, which we will
               carry over to the replacing block. */
            old.close_block();
            old.get_block().unlink();
            return old.get_block();
        }

        private Document finalize_and_process() {
            close_block_parsers(open_block_parsers.size);
            process_inlines();
            return (Document)document_block_parser.get_block();
        }

        private void close_block_parsers(int count) {
            for (int i = 0; i < count; i++) {
                var block_parser = deactivate_block_parser().block_parser;
                do_finalize(block_parser);
                /* Remember for inline parsing. Note that a lot of blocks don't
                   need inline parsing. We could have a separate interface
                   (e.g. BlockParserWithInlines) so that we only have to
                   remember those that actually have inlines to parse. */
                all_block_parsers.add(block_parser);
            }
        }

        private class MatchedBlockParserImpl : GLib.Object,
            MatchedBlockParser {

            private BlockParser matched_block_parser;

            public MatchedBlockParserImpl(
                BlockParser matched_block_parser) {
                this.matched_block_parser = matched_block_parser;
            }

            public BlockParser get_matched_block_parser() {
                return matched_block_parser;
            }

            public SourceLines get_paragraph_lines() {
                if (matched_block_parser is ParagraphParser) {
                    var paragraph_parser = matched_block_parser as
                        ParagraphParser;
                    return paragraph_parser.get_paragraph_lines();
                }
                return SourceLines.empty();
            }
        }

        private class OpenBlockParser : GLib.Object {

            public BlockParser block_parser;
            public int source_index;

            public OpenBlockParser(BlockParser block_parser,
                                   int source_index) {
                this.block_parser = block_parser;
                this.source_index = source_index;
            }
        }

        static construct {
            CORE_FACTORY_TYPES = new LinkedSet<GLib.Type>();
            CORE_FACTORY_TYPES.add_all_array(
                new GLib.Type[] {
                    typeof(BlockQuote),
                    typeof(Heading),
                    typeof(FencedCodeBlock),
                    typeof(HtmlBlock),
                    typeof(ThematicBreak),
                    typeof(ListBlock),
                    typeof(IndentedCodeBlock) });

            NODES_TO_CORE_FACTORIES =
                new Gee.HashMap<GLib.Type, BlockParserFactory>();
            NODES_TO_CORE_FACTORIES[typeof(BlockQuote)] =
                new BlockQuoteParser.Factory();
            NODES_TO_CORE_FACTORIES[typeof(Heading)] =
                new HeadingParser.Factory();
            NODES_TO_CORE_FACTORIES[typeof(FencedCodeBlock)] =
                new FencedCodeBlockParser.Factory();
            NODES_TO_CORE_FACTORIES[typeof(HtmlBlock)] =
                new HtmlBlockParser.Factory();
            NODES_TO_CORE_FACTORIES[typeof(ThematicBreak)] =
                new ThematicBreakParser.Factory();
            NODES_TO_CORE_FACTORIES[typeof(ListBlock)] =
                new ListBlockParser.Factory();
            NODES_TO_CORE_FACTORIES[typeof(IndentedCodeBlock)] =
                new IndentedCodeBlockParser.Factory();
        }
    }
}
