namespace Commonmark {

    public class FencedCodeBlockParser : AbstractBlockParser {

        private FencedCodeBlock block;

        private string first_line;
        private StringBuilder other_lines;

        public FencedCodeBlockParser(char fence_char,
                                     int fence_length,
                                     int fence_indent) {
            block = new FencedCodeBlock();
            other_lines = new StringBuilder();
            block.fence_char = fence_char;
            block.fence_length = fence_length;
            block.fence_indent = fence_indent;
        }

        public override Block get_block() {
            return block;
        }

        public override BlockContinue
            try_continue(ParserState state) {
            int next_non_space = state.get_next_non_space_index();
            int new_index = state.get_index();
            var line = state.get_line().content;
            if (state.get_indent() < Parsing.CODE_BLOCK_INDENT &&
                next_non_space < line.length &&
                line.get(next_non_space) == block.fence_char &&
                is_closing(line, next_non_space)) {
                /* closing fence - we're at end of line, so we can finalize
                   now */
                return BlockContinue.finished();
            } else {
                /* skip optional spaces of fence indent */
                int i = block.fence_indent;
                int length = line.length;
                while (i > 0 && new_index < length &&
                       line.get(new_index) == ' ') {
                    new_index++;
                    i--;
                }
            }
            return BlockContinue.at_index(new_index);
        }

        public override void add_line(SourceLine line) {
            if (first_line == null) {
                first_line = line.content.to_string();
            } else {
                other_lines.append(line.content);
                other_lines.append("\n");
            }
        }

        public override void close_block() {
            /* first line becomes info string */
            block.info = Escaping.unescape_string(
                first_line.strip());
            block.literal = other_lines.str;
        }

        public class Factory : AbstractBlockParserFactory {

            public override BlockStart try_start(
                ParserState state,
                MatchedBlockParser matchedBlockParser) {
                int indent = state.get_indent();
                if (indent >= Parsing.CODE_BLOCK_INDENT) {
                    return BlockStart.none();
                }

                int next_non_space = state.get_next_non_space_index();
                var block_parser = check_opener(state.get_line().content,
                                                next_non_space, indent);
                if (block_parser != null) {
                    return BlockStart.from(block_parser)
                        .at_index(next_non_space +
                                  block_parser.block.fence_length);
                } else {
                    return BlockStart.none();
                }
            }
        }

        /* spec: A code fence is a sequence of at least three consecutive
           backtick characters (`) or tildes (~). (Tildes and backticks cannot
           be mixed.) */
        private static FencedCodeBlockParser? check_opener(string line,
                                                           int index,
                                                           int indent) {
            int backticks = 0;
            int tildes = 0;
            int length = line.length;
            for (int i = index; i < length; i++) {
                var c = line.get(i);
                if (c == '`')
                    backticks++;
                else if (c == '~')
                    tildes++;
                else
                    break;
            }
            if (backticks >= 3 && tildes == 0) {
                /* spec: If the info string comes after a backtick fence, it may
                   not contain any backtick characters. */
                if (Parsing.find('`', line,
                                               index + backticks) != -1) {
                    return null;
                }
                return new FencedCodeBlockParser('`', backticks, indent);
            } else if (tildes >= 3 && backticks == 0) {
                /* spec: Info strings for tilde code blocks can contain
                   backticks and tildes */
                return new FencedCodeBlockParser('~', tildes, indent);
            } else {
                return null;
            }
        }

        /* spec: The content of the code block consists of all subsequent lines,
           until a closing code fence of the same type as the code block began
           with (backticks or tildes), and with at least as many backticks or
           tildes as the opening code fence. */
        private bool is_closing(string line, int index) {
            var fence_char = block.fence_char;
            int fence_length = block.fence_length;
            int fences = Parsing.skip(fence_char, line, index, line.length) -
                index;
            if (fences < fence_length) {
                return false;
            }
            /* spec: The closing code fence [...] may be followed only by
               spaces, which are ignored. */
            int after = Parsing.skip_space_tab(line, index + fences,
                                               line.length);
            return after == line.length;
        }
    }
}
