namespace Commonmark {

    public class HeadingParser : AbstractBlockParser {

        private Heading block;
        private SourceLines content;

        public HeadingParser(int level, SourceLines content) {
            block = new Heading();
            block.level = level;
            this.content = content;
        }

        public override Block get_block() {
            return block;
        }

        public override BlockContinue
            try_continue(ParserState parser_state) {
            /* In both ATX and Setext headings, once we have the heading markup,
               there's nothing more to parse. */
            return BlockContinue.none();
        }

        public override void parse_inlines(InlineParser inline_parser) {
            inline_parser.parse(content, block);
        }

        public class Factory : AbstractBlockParserFactory {

            public override BlockStart try_start(
                ParserState state,
                MatchedBlockParser matched_block_parser) {
                if (state.get_indent() >= Parsing.CODE_BLOCK_INDENT) {
                    return BlockStart.none();
                }

                var line = state.get_line();
                int next_non_space = state.get_next_non_space_index();
                if (line.content.get(next_non_space) == '#') {
                    var atx_heading = get_atx_heading(
                        line.substring(next_non_space, line.content.length));
                    if (atx_heading != null) {
                        return BlockStart.from(
                            atx_heading).at_index(line.content.length);
                    }
                }

                int setext_heading_level =
                    get_setext_heading_level(line.content, next_non_space);
                if (setext_heading_level > 0) {
                    var paragraph = matched_block_parser.get_paragraph_lines();
                    if (!paragraph.is_empty) {
                        return BlockStart.from(
                            new HeadingParser(setext_heading_level, paragraph))
                            .at_index(line.content.length)
                            .replace_active_block_parser();
                    }
                }

                return BlockStart.none();
            }
        }

        /* spec: An ATX heading consists of a string of characters, parsed as
           inline content, between an opening sequence of 1–6 unescaped #
           characters and an optional closing sequence of any number of
           unescaped # characters. The opening sequence of # characters must be
           followed by a space or by the end of line. The optional closing
           sequence of #s must be preceded by a space and may be followed by
           spaces only. */
        private static HeadingParser? get_atx_heading(SourceLine line) {
            var scanner = Scanner.from(SourceLines.from_source_line(line));
            int level = scanner.match_multiple('#');

            if (level == 0 || level > 6)
                return null;

            if (!scanner.has_next()) {
                /* End of line after markers is an empty heading */
                return new HeadingParser(level, SourceLines.empty());
            }

            unichar next = scanner.peek();
            if (!(next == ' ' || next == '\t')) {
                return null;
            }

            scanner.whitespace();
            var start = scanner.position();
            var end = start;
            bool hash_can_end = true;

            while (scanner.has_next()) {
                unichar c = scanner.peek();
                switch (c) {
                case '#':
                    if (hash_can_end) {
                        scanner.match_multiple('#');
                        int whitespace = scanner.whitespace();
                        /* If there's other characters, the hashes and spaces
                           were part of the heading */
                        if (scanner.has_next()) {
                            end = scanner.position();
                        }
                        hash_can_end = whitespace > 0;
                    } else {
                        scanner.next();
                        end = scanner.position();
                    }
                    break;
                case ' ':
                case '\t':
                    hash_can_end = true;
                    scanner.next();
                    break;
                default:
                    hash_can_end = false;
                    scanner.next();
                    end = scanner.position();
                    break;
                }
            }

            var source = scanner.get_source(start, end);
            var content = source.get_content();
            if (content == "") {
                return new HeadingParser(level, SourceLines.empty());
            }
            return new HeadingParser(level, source);
        }

        /* spec: A setext heading underline is a sequence of = characters or a
           sequence of - characters, with no more than 3 spaces indentation and
           any number of trailing spaces. */
        private static int get_setext_heading_level(string line,
                                                    int index) {
            switch (line.get(index)) {
            case '=':
                if (is_setext_heading_rest(line, index + 1, '=')) {
                    return 1;
                }
                return 2;
            case '-':
                if (is_setext_heading_rest(line, index + 1, '-')) {
                    return 2;
                }
                return 0;
            }
            return 0;
        }

        private static bool is_setext_heading_rest(string line,
                                                   int index,
                                                   char marker) {
            int after_marker = Parsing.skip(marker, line, index, line.length);
            int after_space = Parsing.skip_space_tab(line, after_marker,
                                                     line.length);
            return after_space >= line.length;
        }
    }
}
