namespace Commonmark {

    /**
     * Attempt to parse inline HTML.
     */
    public class HtmlInlineParser : GLib.Object, InlineContentParser {

        private static AsciiMatcher ASCII_LETTER;
        private static AsciiMatcher TAG_NAME_START;
        private static AsciiMatcher TAG_NAME_CONTINUE;
        private static AsciiMatcher ATTRIBUTE_START;
        private static AsciiMatcher ATTRIBUTE_CONTINUE;
        private static AsciiMatcher ATTRIBUTE_VALUE_END;

        public ParsedInline try_parse(InlineParserState inline_parser_state) {
            Scanner scanner = inline_parser_state.scanner();
            Position start = scanner.position();
            /* Skip over `<` */
            scanner.next();

            unichar c = scanner.peek();
            if (TAG_NAME_START.matches(c)) {
                if (try_open_tag(scanner)) {
                    return html_inline(start, scanner);
                }
            } else if (c == '/') {
                if (try_closing_start(scanner)) {
                    return html_inline(start, scanner);
                }
            } else if (c == '?') {
                if (try_processing_instruction(scanner)) {
                    return html_inline(start, scanner);
                }
            } else if (c == '!') {
                /* comment, declaration or CDATA */
                scanner.next();
                c = scanner.peek();
                if (c == '-') {
                    if (try_comment(scanner)) {
                        return html_inline(start, scanner);
                    }
                } else if (c == '[') {
                    if (try_cdata(scanner)) {
                        return html_inline(start, scanner);
                    }
                } else if (ASCII_LETTER.matches(c)) {
                    if (try_declaration(scanner)) {
                        return html_inline(start, scanner);
                    }
                }
            }

            return ParsedInline.none();
        }

        private static ParsedInline html_inline(Position start, Scanner scanner) {
            string text = scanner.get_source(start, scanner.position()).get_content();
            var node = new HtmlInline();
            node.literal = text;
            return ParsedInline.from(node, scanner.position());
        }

        private static bool try_open_tag(Scanner scanner) {
            /* spec: An open tag consists of a < character, a tag name, zero or
               more attributes, optional whitespace, an optional / character,
               and a > character. */
            scanner.next();
            scanner.match(TAG_NAME_CONTINUE);
            bool whitespace = scanner.whitespace() >= 1;
            /* spec: An attribute consists of whitespace, an attribute name, and
               an optional attribute value specification. */
            while (whitespace && scanner.match(ATTRIBUTE_START) >= 1) {
                scanner.match(ATTRIBUTE_CONTINUE);
                /* spec: An attribute value specification consists of optional
                   whitespace, a = character, optional whitespace, and an
                   attribute value.*/
                whitespace = scanner.whitespace() >= 1;
                if (scanner.next_char('=')) {
                    scanner.whitespace();
                    unichar value_start = scanner.peek();
                    if (value_start == '\'') {
                        scanner.next();
                        if (scanner.find('\'') < 0) {
                            return false;
                        }
                        scanner.next();
                    } else if (value_start == '"') {
                        scanner.next();
                        if (scanner.find('"') < 0) {
                            return false;
                        }
                        scanner.next();
                    } else {
                        if (scanner.find_matcher(ATTRIBUTE_VALUE_END) <= 0) {
                            return false;
                        }
                    }

                    /* Whitespace is required between attributes */
                    whitespace = scanner.whitespace() >= 1;
                }
            }

            scanner.next_char('/');
            return scanner.next_char('>');
        }

        private static bool try_closing_start(Scanner scanner) {
            /* spec: A closing tag consists of the string </, a tag name,
               optional whitespace, and the character >. */
            scanner.next();
            if (scanner.match(TAG_NAME_START) >= 1) {
                scanner.match(TAG_NAME_CONTINUE);
                scanner.whitespace();
                return scanner.next_char('>');
            }
            return false;
        }

        private static bool try_processing_instruction(Scanner scanner) {
            /* spec: A processing instruction consists of the string <?, a
               string of characters not including the string ?>, and the string
               ?>. */
            scanner.next();
            while (scanner.find('?') > 0) {
                scanner.next();
                if (scanner.next_char('>')) {
                    return true;
                }
            }
            return false;
        }

        private static bool try_comment(Scanner scanner) {
            /* spec: An HTML comment consists of <!-- + text + -->, where text
               does not start with > or ->, does not end with -, and does not
               contain --. (See the HTML5 spec.) */

            /* Skip first `-` */
            scanner.next();
            if (!scanner.next_char('-')) {
                return false;
            }

            if (scanner.next_char('>') || scanner.next_content("->")) {
                return false;
            }

            while (scanner.find('-') >= 0) {
                if (scanner.next_content("--")) {
                    return scanner.next_char('>');
                } else {
                    scanner.next();
                }
            }

            return false;
        }

        private static bool try_cdata(Scanner scanner) {
            /* spec: A CDATA section consists of the string <![CDATA[, a string
               of characters not including the string ]]>, and the string
               ]]>. */

            /* Skip `[` */
            scanner.next();

            if (scanner.next_content("CDATA[")) {
                while (scanner.find(']') >= 0) {
                    if (scanner.next_content("]]>")) {
                        return true;
                    } else {
                        scanner.next();
                    }
                }
            }

            return false;
        }

        private static bool try_declaration(Scanner scanner) {
            /* spec: A declaration consists of the string <!, an ASCII letter,
               zero or more characters not including the character >, and the
               character >. */
            scanner.match(ASCII_LETTER);
            if (scanner.whitespace() <= 0) {
                return false;
            }
            if (scanner.find('>') >= 0) {
                scanner.next();
                return true;
            }
            return false;
        }

        static construct {
            ASCII_LETTER = AsciiMatcher.builder().range('A', 'Z')
                .range('a', 'z').build();

            /* spec: A tag name consists of an ASCII letter followed by zero or
               more ASCII letters, digits, or hyphens (-). */
            TAG_NAME_START = ASCII_LETTER;
            TAG_NAME_CONTINUE = TAG_NAME_START.new_builder().range('0', '9')
                .c('-').build();

            /* spec: An attribute name consists of an ASCII letter, _, or :,
               followed by zero or more ASCII letters, digits, _, ., :, or
               -. (Note: This is the XML specification restricted to
               ASCII. HTML5 is laxer.) */
            ATTRIBUTE_START = ASCII_LETTER.new_builder().c('_').c(':').build();
            ATTRIBUTE_CONTINUE = ATTRIBUTE_START.new_builder().range('0', '9')
                .c('.').c('-').build();
            /* spec: An unquoted attribute value is a nonempty string of
             * characters not including whitespace, ", ', =, <, >, or `. */
            ATTRIBUTE_VALUE_END = AsciiMatcher.builder()
                .c(' ').c('\t').c('\n').c('\x0b').c('\f').c('\r')
                .c('"').c('\'').c('=').c('<').c('>').c('`')
                .build();
        }
    }
}
