namespace Commonmark {

    public class InlineParserImpl : GLib.Object, InlineParser,
        InlineParserState {

        private BitSet special_characters;
        private Gee.Map<unichar, DelimiterProcessor> delimiter_processors;
        private InlineParserContext context;
        private Gee.Map<unichar, Gee.List<InlineContentParser>> inline_parsers;

        private Scanner _scanner;
        private bool include_source_spans;
        private int trailing_spaces;

        /**
         * Top delimiter (emphasis, strong emphasis or custom
         * emphasis). (Brackets are on a separate stack, different from the
         * algorithm described in the spec.)
         */
        private Delimiter last_delimiter;

        /**
         * Top opening bracket (''['' or ''![)'').
         */
        private Bracket last_bracket;

        public InlineParserImpl(InlineParserContext inline_parser_context) {
            this.delimiter_processors =
                calculate_delimiter_processors(
                    inline_parser_context.get_custom_delimiter_processors());

            this.context = inline_parser_context;
            this.inline_parsers =
                new Gee.HashMap<unichar,
                Gee.List<InlineContentParser>>();
            this.inline_parsers['\\'] =
                singleton_list(new BackslashInlineParser());
            this.inline_parsers['`'] =
                singleton_list(new BackticksInlineParser());
            this.inline_parsers['&'] =
                singleton_list(new EntityInlineParser());
            this.inline_parsers['<'] =
                doubleton_list(new AutolinkInlineParser(),
                               new HtmlInlineParser());

            this.special_characters =
                calculate_special_characters(this.delimiter_processors.keys,
                                             inline_parsers.keys);
        }

        private Gee.List<InlineContentParser>
            singleton_list(InlineContentParser parser) {
            var list = new Gee.LinkedList<InlineContentParser>();
            list.add(parser);
            return list;
        }

        private Gee.List<InlineContentParser>
            doubleton_list(InlineContentParser parser1,
                           InlineContentParser parser2) {
            var list = new Gee.LinkedList<InlineContentParser>();
            list.add(parser1);
            list.add(parser2);
            return list;
        }

        public static BitSet
            calculate_special_characters(Gee.Set<unichar> delimiter_characters,
                                         Gee.Set<unichar> characters) {
            var bit_set = new BitSet();
            foreach (var c in delimiter_characters) {
                bit_set.set_bit((int)c);
            }
            foreach (var c in characters) {
                bit_set.set_bit((int)c);
            }
            bit_set.set_bit('[');
            bit_set.set_bit(']');
            bit_set.set_bit('!');
            bit_set.set_bit('\n');
            return bit_set;
        }

        public static Gee.Map<unichar, DelimiterProcessor>
            calculate_delimiter_processors(
                Gee.List<DelimiterProcessor>
                delimiter_processors) {
            var map = new Gee.HashMap<unichar,
                DelimiterProcessor>();
            var list = new Gee.LinkedList<
                DelimiterProcessor>();
            list.add(new AsteriskDelimiterProcessor());
            list.add(new UnderscoreDelimiterProcessor());
            add_delimiter_processors(list, map);
            add_delimiter_processors(delimiter_processors, map);
            return map;
        }

        public Scanner scanner() {
            return _scanner;
        }

        private static void add_delimiter_processors(
            Gee.Iterable<DelimiterProcessor> delimiter_processors,
            Gee.Map<unichar, DelimiterProcessor> map) {
            foreach (var delimiter_processor in delimiter_processors) {
                var opening = delimiter_processor.get_opening_character();
                var closing = delimiter_processor.get_closing_character();
                if (opening == closing) {
                    DelimiterProcessor old = map[opening];
                    if (old != null &&
                        old.get_opening_character() ==
                        old.get_closing_character()) {
                        StaggeredDelimiterProcessor s;
                        if (old is StaggeredDelimiterProcessor) {
                            s = old as StaggeredDelimiterProcessor;
                        } else {
                            s = new StaggeredDelimiterProcessor(opening);
                            s.add(old);
                        }
                        s.add(delimiter_processor);
                        map[opening] = s;
                    } else {
                        add_delimiter_processor_for_char(
                            opening, delimiter_processor, map);
                    }
                } else {
                    add_delimiter_processor_for_char(
                        opening, delimiter_processor, map);
                    add_delimiter_processor_for_char(
                        closing, delimiter_processor, map);
                }
            }
        }

        private static void
            add_delimiter_processor_for_char(
            unichar delimiter_char, DelimiterProcessor to_add,
            Gee.Map<unichar,
            DelimiterProcessor> delimiter_processors) {
            if (delimiter_processors.has_key(delimiter_char))
                GLib.error("Delimiter processor conflict with delimiter char '%c'",
                           (int)delimiter_char);
            delimiter_processors[delimiter_char] = to_add;
        }

        /**
         * Parse content in block into inline children, appending them to the
         * block node.
         */
        public void parse(SourceLines lines, Node block) {
            reset(lines);

            while (true) {
                Gee.List<Node> nodes = parse_inline();
                if (nodes != null) {
                    foreach (var node in nodes) {
                        block.append_child(node);
                    }
                } else {
                    break;
                }
            }

            process_delimiters(null);
            merge_child_text_nodes(block);
        }

        void reset(SourceLines lines) {
            this._scanner = Scanner.from(lines);
            this.include_source_spans = !lines.get_source_spans().is_empty;
            this.trailing_spaces = 0;
            this.last_delimiter = null;
            this.last_bracket = null;
        }

        private Text text(SourceLines source_lines) {
            var text = new Text.from_literal(source_lines.get_content());
            text.source_spans = source_lines.get_source_spans();
            return text;
        }

        private Gee.List<Node> singleton_list_node(Node node) {
            var list = new Gee.LinkedList<Node>();
            list.add(node);
            return list;
        }

        /**
         * Parse the next inline element in subject, advancing our position.
         * On success, return the new inline node.
         * On failure, return null.
         */
        private Gee.List<Node>? parse_inline() {
            var c = _scanner.peek();

            switch (c) {
                case '[':
                    return singleton_list_node(parse_open_bracket());
                case '!':
                    return singleton_list_node(parse_bang());
                case ']':
                    return singleton_list_node(parse_close_bracket());
                case '\n':
                    return singleton_list_node(parse_line_break());
                case Scanner.END:
                    return null;
            }

            /* No inline parser, delimiter or other special handling. */
            if (!special_characters.get_bit((int)c))
                return singleton_list_node(parse_text());

            Gee.List<InlineContentParser> inline_parsers =
                this.inline_parsers[c];
            if (inline_parsers != null) {
                var position = _scanner.position();
                foreach (var inline_parser in inline_parsers) {
                    var parsed_inline = inline_parser.try_parse(this);
                    if (parsed_inline is ParsedInlineImpl) {
                        var parsed_inline_impl = parsed_inline as
                            ParsedInlineImpl;
                        var node = parsed_inline_impl.node;
                        _scanner.set_position(parsed_inline_impl.position);
                        if (include_source_spans &&
                            node.source_spans.is_empty) {
                            node.source_spans = _scanner.get_source(
                                position, _scanner.position())
                                .get_source_spans();
                        }
                        return singleton_list_node(node);
                    } else {
                        /* Reset position */
                        _scanner.set_position(position);
                    }
                }
            }

            var delimiter_processor = delimiter_processors[c];
            if (delimiter_processor != null) {
                var nodes = parse_delimiters(delimiter_processor, c);
                if (nodes != null)
                    return nodes;
            }

            /* If we get here, even for a special/delimiter character, we will
               just treat it as text. */
            return singleton_list_node(parse_text());
        }

        /**
         * Attempt to parse delimiters like emphasis, strong emphasis or custom
         * delimiters.
         */
        private Gee.List<Node>? parse_delimiters(
            DelimiterProcessor delimiter_processor,
            unichar delimiter_char) {
            var res = scan_delimiters(delimiter_processor, delimiter_char);
            if (res == null)
                return null;

            var characters = res.characters;

            /* Add entry to stack for this opener */
            last_delimiter = new Delimiter(characters, delimiter_char,
                                           res.can_open, res.can_close,
                                           last_delimiter);
            if (last_delimiter.previous != null)
                last_delimiter.previous.next = last_delimiter;

            return characters;
        }

        /**
         * Add open bracket to delimiter stack and add a text node to block's
         * children.
         */
        private Node parse_open_bracket() {
            var start = _scanner.position();
            _scanner.next();
            var content_position = _scanner.position();

            var node = text(_scanner.get_source(start, content_position));

            /* Add entry to stack for this opener */
            add_bracket(Bracket.link(node, start, content_position,
                                     last_bracket, last_delimiter));

            return node;
        }

        /**
         * If next character is [, and ! delimiter to delimiter stack and add a
         * text node to block's children.  Otherwise just add a text node.
         */
        private Node parse_bang() {
            var start = _scanner.position();
            _scanner.next();
            if (_scanner.next_char('[')) {
                var content_position = _scanner.position();
                var node = text(_scanner.get_source(start, content_position));

                /* Add entry to stack for this opener */
                add_bracket(Bracket.image(node, start, content_position,
                                          last_bracket, last_delimiter));
                return node;
            } else {
                return text(_scanner.get_source(start, _scanner.position()));
            }
        }

        /**
         * Try to match close bracket against an opening in the delimiter
         * stack. Return either a link or image, or a plain [ character. If
         * there is a matching delimiter, remove it from the delimiter stack.
         */
        private Node parse_close_bracket() {
            var before_close = _scanner.position();
            _scanner.next();
            var after_close = _scanner.position();

            /* Get previous `[` or `![` */
            var opener = last_bracket;
            if (opener == null)
                /* No matching opener, just return a literal. */
                return text(_scanner.get_source(before_close, after_close));

            if (!opener.allowed) {
                /* Matching opener but it's not allowed, just return a
                   literal. */
                remove_last_bracket();
                return text(_scanner.get_source(before_close, after_close));
            }

            /* Check to see if we have a link/image */
            string dest = null;
            string title = null;

            /* Maybe a inline link like `[foo](/uri "title")` */
            if (_scanner.next_char('(')) {
                _scanner.whitespace();
                dest = parse_link_destination(_scanner);
                if (dest == null) {
                    _scanner.set_position(after_close);
                } else {
                    int whitespace = _scanner.whitespace();
                    /* title needs a whitespace before */
                    if (whitespace >= 1) {
                        title = parse_link_title(_scanner);
                        _scanner.whitespace();
                    }
                    if (!_scanner.next_char(')')) {
                        /* Don't have a closing `)`, so it's not a destination
                           and title -> reset.  Note that something like
                           `[foo](` could be valid, `(` will just be text.*/
                        _scanner.set_position(after_close);
                        dest = null;
                        title = null;
                    }
                }
            }

            /* Maybe a reference link like `[foo][bar]`, `[foo][]` or `[foo]`.
               Note that even `[foo](` could be a valid link if there's a
               reference, which is why this is not just an `else` here. */
            if (dest == null) {
                /* See if there's a link label like `[bar]` or `[]` */
                var reference = parse_link_label(_scanner);
                if (reference == null)
                    _scanner.set_position(after_close);
                if ((reference == null || reference.length == 0) &&
                    !opener.bracket_after) {
                    /* If the second label is empty `[foo][]` or missing
                       `[foo]`, then the first label is the reference.  But it
                       can only be a reference when there's no (unescaped)
                       bracket in it.  If there is, we don't even need to try to
                       look up the reference. This is an optimization. */
                    reference = _scanner.get_source(
                        opener.content_position, before_close).get_content();
                }

                if (reference != null) {
                    var definition =
                        context.get_link_reference_definition(reference);
                    if (definition != null) {
                        dest = definition.destination;
                        title = definition.title;
                    }
                }
            }

            if (dest != null) {
                /* If we got here, we have a link or image */
                Node link_or_image;
                if (opener._image)
                    link_or_image = new Image.full(dest, title);
                else
                    link_or_image = new Link.full(dest, title);

                /* Add all nodes between the opening bracket and now (closing
                   bracket) as child nodes of the link */
                var node = opener.node.next;
                while (node != null) {
                    var next = node.next;
                    link_or_image.append_child(node);
                    node = next;
                }

                if (include_source_spans)
                    link_or_image.source_spans =
                        _scanner.get_source(
                            opener.marker_position, _scanner.position())
                        .get_source_spans();

                /* Process delimiters such as emphasis inside link/image */
                process_delimiters(opener.previous_delimiter);
                merge_child_text_nodes(link_or_image);
                /* We don't need the corresponding text node anymore, we turned
                   it into a link/image node */
                opener.node.unlink();
                remove_last_bracket();

                /* Links within links are not allowed. We found this link, so
                   there can be no other link around it. */
                if (!opener._image) {
                    var bracket = last_bracket;
                    while (bracket != null) {
                        if (!bracket._image) {
                            /* Disallow link opener. It will still get matched,
                               but will not result in a link. */
                            bracket.allowed = false;
                        }
                        bracket = bracket.previous;
                    }
                }
                return link_or_image;
            } else {
                /* No link or image, parse just the bracket as text and
                   continue */
                remove_last_bracket();
                _scanner.set_position(after_close);
                return text(_scanner.get_source(before_close, after_close));
            }
        }

        private void add_bracket(Bracket bracket) {
            if (last_bracket != null) {
                last_bracket.bracket_after = true;
            }
            last_bracket = bracket;
        }

        private void remove_last_bracket() {
            last_bracket = last_bracket.previous;
        }

        /**
         * Attempt to parse link destination, returning the string or null if no match.
         */
        private string? parse_link_destination(Scanner _scanner) {
            var delimiter = _scanner.peek();
            var start = _scanner.position();
            if (!LinkScanner.scan_link_destination(_scanner))
                return null;

            string dest;
            if (delimiter == '<') {
                /* chop off surrounding <..>: */
                var raw_destination = _scanner
                    .get_source(start, _scanner.position()).get_content();
                dest = raw_destination.slice(1, raw_destination.length - 1);
            } else {
                dest = _scanner.get_source(start, _scanner.position())
                    .get_content();
            }

            return Escaping.unescape_string(dest);
        }

        /**
         * Attempt to parse link title (sans quotes), returning the string or
         * null if no match.
         */
        private string? parse_link_title(Scanner scanner) {
            var start = scanner.position();
            if (!LinkScanner.scan_link_title(scanner))
                return null;

            /* chop off ', " or parens */
            string raw_title = scanner.get_source(start, scanner.position())
                .get_content();
            string title = raw_title.slice(1, raw_title.length - 1);
            return Escaping.unescape_string(title);
        }

        /**
         * Attempt to parse a link label, returning the label between the
         * brackets or null.
         */
        public string? parse_link_label(Scanner scanner) {
            if (!scanner.next_char('['))
                return null;

            var start = scanner.position();
            if (!LinkScanner.scan_link_label_content(scanner))
                return null;
            var end = scanner.position();

            if (!scanner.next_char(']'))
                return null;

            string content = scanner.get_source(start, end).get_content();
            /* spec: A link label can have at most 999 characters inside the
               square brackets. */
            if (content.length > 999)
                return null;

            return content;
        }

        private Node parse_line_break() {
            _scanner.next();

            if (trailing_spaces >= 2) {
                return new HardLineBreak();
            } else {
                return new SoftLineBreak();
            }
        }

        /**
         * Parse the next character as plain text, and possibly more if the
         * following characters are non-special.
         */
        private Node parse_text() {
            var start = _scanner.position();
            _scanner.next();
            unichar c;
            while (true) {
                c = _scanner.peek();
                if (c == Scanner.END ||
                    special_characters.get_bit((int)c))
                    break;
                _scanner.next();
            }

            var source = _scanner.get_source(start, _scanner.position());
            var content = source.get_content();

            if (c == '\n') {
                /* We parsed until the end of the line. Trim any trailing spaces
                   and remember them (for hard line breaks). */
                int end = Parsing
                    .skip_backwards(' ', content, content.length - 1, 0) + 1;
                trailing_spaces = content.length - end;
                content = content.slice(0, end);
            } else if (c == Scanner.END) {
                /* For the last line, both tabs and spaces are trimmed for some
                 * reason (checked with commonmark.js). */
                int end = Parsing
                    .skip_space_tab_backwards(
                        content, content.length - 1, 0) + 1;
                content = content.slice(0, end);
            }

            var text = new Text.from_literal(content);
            text.source_spans = source.get_source_spans();
            return text;
        }

        /**
         * Scan a sequence of characters with code delimiterChar, and return
         * information about the number of delimiters and whether they are
         * positioned such that they can open and/or close emphasis or strong
         * emphasis.
         *
         * @return information about delimiter run, or {@code null}
         */
        private DelimiterData? scan_delimiters(
            DelimiterProcessor delimiter_processor,
            unichar delimiter_char) {
            int before = _scanner.peek_previous_code_point();
            var start = _scanner.position();

            /* Quick check to see if we have enough delimiters. */
            int delimiter_count = _scanner.match_multiple(delimiter_char);
            if (delimiter_count < delimiter_processor.get_min_length()) {
                _scanner.set_position(start);
                return null;
            }

            /* We do have enough, extract a text node for each delimiter
               character. */
            var delimiters = new Gee.ArrayList<Text>();
            _scanner.set_position(start);
            var position_before = start;
            while (_scanner.next_char(delimiter_char)) {
                delimiters.add(text(_scanner.get_source(position_before,
                                                        _scanner.position())));
                position_before = _scanner.position();
            }

            int after = _scanner.peek_code_point();

            /* We could be more lazy here, in most cases we don't need to do
              every match case. */
            bool before_is_punctuation =
                before == Scanner.END ||
                Parsing.is_punctuation_code_point(before);
            bool before_is_whitespace =
                before == Scanner.END ||
                Parsing.is_whitespace_code_point(before);
            bool after_is_punctuation =
                after == Scanner.END ||
                Parsing.is_punctuation_code_point(after);
            bool after_is_whitespace =
                after == Scanner.END ||
                Parsing.is_whitespace_code_point(after);

            bool left_flanking = !after_is_whitespace &&
                    (!after_is_punctuation || before_is_whitespace ||
                     before_is_punctuation);
            bool right_flanking = !before_is_whitespace &&
                    (!before_is_punctuation || after_is_whitespace ||
                     after_is_punctuation);
            bool can_open;
            bool can_close;
            if (delimiter_char == '_') {
                can_open = left_flanking && (!right_flanking ||
                                             before_is_punctuation);
                can_close = right_flanking && (!left_flanking ||
                                               after_is_punctuation);
            } else {
                can_open = left_flanking && delimiter_char ==
                    delimiter_processor.get_opening_character();
                can_close = right_flanking && delimiter_char ==
                    delimiter_processor.get_closing_character();
            }

            return new DelimiterData(delimiters, can_open, can_close);
        }

        private void process_delimiters(Delimiter? stack_bottom) {

            var openers_bottom = new Gee.HashMap<unichar, Delimiter>();

            /* find first closer above stackBottom: */
            Delimiter closer = last_delimiter;
            while (closer != null && closer.previous != stack_bottom)
                closer = closer.previous;
            /* move forward, looking for closers, and handling each */
            while (closer != null) {
                unichar delimiter_char = closer.delimiter_char;

                var delimiter_processor = delimiter_processors.get(delimiter_char);
                if (!closer.can_close() || delimiter_processor == null) {
                    closer = closer.next;
                    continue;
                }

                unichar opening_delimiter_char =
                    delimiter_processor.get_opening_character();

                /* Found delimiter closer. Now look back for first matching
                   opener. */
                int used_delims = 0;
                bool opener_found = false;
                bool potential_opener_found = false;
                Delimiter opener = closer.previous;
                while (opener != null &&
                       opener != stack_bottom &&
                       opener != openers_bottom.get(delimiter_char)) {
                    if (opener.can_open() &&
                        opener.delimiter_char == opening_delimiter_char) {
                        potential_opener_found = true;
                        used_delims =
                            delimiter_processor.process(opener, closer);
                         if (used_delims > 0) {
                             opener_found = true;
                             break;
                         }
                    }
                    opener = opener.previous;
                }

                if (!opener_found) {
                    if (!potential_opener_found) {
                        /* Set lower bound for future searches for openers.
                           Only do this when we didn't even have a potential
                           opener (one that matches the character and can open).
                           If an opener was rejected because of the number of
                           delimiters (e.g. because of the "multiple of 3"
                           rule), we want to consider it next time because the
                           number of delimiters can change as we continue
                           processing. */
                        openers_bottom[delimiter_char] = closer.previous;
                        if (!closer.can_open()) {
                            /* We can remove a closer that can't be an opener,
                               once we've seen there's no matching opener: */
                            remove_delimiter_keep_node(closer);
                        }
                    }
                    closer = closer.next;
                    continue;
                }

                /* Remove number of used delimiters nodes. */
                for (int i = 0; i < used_delims; i++) {
                    var delimiter =
                        opener.characters.remove_at(opener.characters.size - 1);
                    delimiter.unlink();
                }
                for (int i = 0; i < used_delims; i++) {
                    var delimiter = closer.characters.remove_at(0);
                    delimiter.unlink();
                }

                remove_delimiters_between(opener, closer);

                /* No delimiter characters left to process, so we can remove
                   delimiter and the now empty node. */
                if (opener.length() == 0)
                    remove_delimiter_and_nodes(opener);

                if (closer.length() == 0) {
                    var next = closer.next;
                    remove_delimiter_and_nodes(closer);
                    closer = next;
                }
            }

            /* remove all delimiters */
            while (last_delimiter != null && last_delimiter != stack_bottom)
                remove_delimiter_keep_node(last_delimiter);
        }

        private void remove_delimiters_between(Delimiter opener,
                                               Delimiter closer) {
            var delimiter = closer.previous;
            while (delimiter != null && delimiter != opener) {
                var previous_delimiter = delimiter.previous;
                remove_delimiter_keep_node(delimiter);
                delimiter = previous_delimiter;
            }
        }

        /**
         * Remove the delimiter and the corresponding text node. For used
         * delimiters, e.g. `*` in `*foo*`.
         */
        private void remove_delimiter_and_nodes(Delimiter delim) {
            remove_delimiter(delim);
        }

        /**
         * Remove the delimiter but keep the corresponding node as text. For
         * unused delimiters such as `_` in `foo_bar`.
         */
        private void remove_delimiter_keep_node(Delimiter delim) {
            remove_delimiter(delim);
        }

        private void remove_delimiter(Delimiter delim) {
            if (delim.previous != null)
                delim.previous.next = delim.next;
            if (delim.next == null)
                /* top of stack */
                last_delimiter = delim.previous;
            else
                delim.next.previous = delim.previous;
        }

        private void merge_child_text_nodes(Node node) {
            /* No children, no need for merging */
            if (node.first_child == null) {
                return;
            }

            merge_text_nodes_inclusive(node.first_child, node.last_child);
        }

        private void merge_text_nodes_inclusive(Node from_node,
                                                Node to_node) {
            Text first = null;
            Text last = null;
            int length = 0;

            var node = from_node;
            while (node != null) {
                if (node is Text) {
                    Text text = (Text)node;
                    if (first == null)
                        first = text;
                    length += text.literal.length;
                    last = text;
                } else {
                    merge_if_needed(first, last, length);
                    first = null;
                    last = null;
                    length = 0;

                    merge_child_text_nodes(node);
                }
                if (node == to_node)
                    break;
                node = node.next;
            }

            merge_if_needed(first, last, length);
        }

        private void merge_if_needed(Text? first,
                                     Text? last,
                                     int text_length) {
            if (first != null && last != null && first != last) {
                var sb = new StringBuilder();
                sb.append(first.literal);
                SourceSpans source_spans = null;
                if (include_source_spans) {
                    source_spans = new SourceSpans();
                    source_spans.add_all(first.source_spans);
                }
                var node = first.next;
                var stop = last.next;
                while (node != stop) {
                    sb.append(((Text)node).literal);
                    if (source_spans != null)
                        source_spans.add_all(node.source_spans);

                    var unlink = node;
                    node = node.next;
                    unlink.unlink();
                }
                var literal = sb.str;
                first.literal = literal;
                if (source_spans != null)
                    first.source_spans = source_spans.get_source_spans();
            }
        }

        private class DelimiterData {

            public Gee.List<Text> characters;
            public  bool can_close;
            public  bool can_open;

            public DelimiterData(Gee.List<Text> characters,
                                 bool can_open, bool can_close) {
                this.characters = characters;
                this.can_open = can_open;
                this.can_close = can_close;
            }
        }
    }
}
