namespace Commonmark {

    /**
     * Parser for link reference definitions at the beginning of a paragraph.
     * [[https://spec.commonmark.org/0.29/#link-reference-definition]]
     */
    public class LinkReferenceDefinitionParser {

        private State state = State.START_DEFINITION;

        private Gee.List<SourceLine> paragraph_lines =
            new Gee.ArrayList<SourceLine>();
        private Gee.List<LinkReferenceDefinition> definitions =
            new Gee.ArrayList<LinkReferenceDefinition>();
        private Gee.List<SourceSpan> source_spans =
            new Gee.ArrayList<SourceSpan>();

        private StringBuilder _label;
        private string _destination;
        private unichar title_delimiter;
        private StringBuilder _title;
        private bool reference_valid;

        public void parse(SourceLine line) {
            paragraph_lines.add(line);
            if (state == State.PARAGRAPH) {
                /* We're in a paragraph now. Link reference definitions can only
                   appear at the beginning, so once we're in a paragraph,
                   there's no going back.*/
                return;
            }

            var scanner = Scanner.from(SourceLines.from_source_line(line));
            while (scanner.has_next()) {
                bool success;
                switch (state) {
                case START_DEFINITION:
                    success = start_definition(scanner);
                    break;
                case LABEL:
                    success = label(scanner);
                    break;
                case DESTINATION:
                    success = destination(scanner);
                    break;
                case START_TITLE:
                    success = start_title(scanner);
                    break;
                case TITLE:
                    success = title(scanner);
                    break;
                default:
                    GLib.error("Unknown parsing state: %d", state);
                }
                /* Parsing failed, which means we fall back to treating text as
                 * a paragraph. */
                if (!success) {
                    state = State.PARAGRAPH;
                    return;
                }
            }
        }

        public void add_source_span(SourceSpan source_span) {
            source_spans.add(source_span);
        }

        /**
         * @return the lines that are normal paragraph content, without newlines
         */
        public SourceLines get_paragraph_lines() {
            return SourceLines.from_source_lines(paragraph_lines);
        }

        public Gee.List<SourceSpan> get_paragraph_source_spans() {
            return source_spans;
        }

        public Gee.List<LinkReferenceDefinition> get_definitions() {
            finish_reference();
            return definitions;
        }

        public State get_state() {
            return state;
        }

        private bool start_definition(Scanner scanner) {
            scanner.whitespace();
            if (!scanner.next_char('['))
                return false;

            state = State.LABEL;
            _label = new StringBuilder();

            if (!scanner.has_next())
                _label.append_c('\n');
            return true;
        }

        private bool label(Scanner scanner) {
            var start = scanner.position();
            if (!LinkScanner.scan_link_label_content(scanner))
                return false;

            _label.append(scanner.get_source(start, scanner.position())
                         .get_content());

            if (!scanner.has_next()) {
                /* label might continue on next line */
                _label.append("\n");
                return true;
            } else if (scanner.next_char(']')) {
                /* end of label */
                if (!scanner.next_char(':'))
                    return false;

                /* spec: A link label can have at most 999 characters inside the
                   square brackets. */
                if (_label.len > 999)
                    return false;

                var normalized_label =
                    Escaping.normalize_label_content(_label.str);
                if (normalized_label == "")
                    return false;

                state = State.DESTINATION;

                scanner.whitespace();
                return true;
            } else {
                return false;
            }
        }

        private bool destination(Scanner scanner) {
            scanner.whitespace();
            var start = scanner.position();
            if (!LinkScanner.scan_link_destination(scanner))
                return false;

            var raw_destination = scanner.get_source(start, scanner.position())
                .get_content();
            _destination = raw_destination.has_prefix("<") ?
                raw_destination.slice(1, raw_destination.length - 1) :
                raw_destination;

            int whitespace = scanner.whitespace();
            if (!scanner.has_next()) {
                /* Destination was at end of line, so this is a valid reference
                   for sure (and maybe a title). If not at end of line, wait for
                   title to be valid first. */
                reference_valid = true;
                paragraph_lines.clear();
            } else if (whitespace == 0) {
                /* spec: The title must be separated from the link destination
                   by whitespace */
                return false;
            }

            state = State.START_TITLE;
            return true;
        }

        private bool start_title(Scanner scanner) {
            scanner.whitespace();
            if (!scanner.has_next()) {
                state = State.START_DEFINITION;
                return true;
            }

            title_delimiter = '\0';
            unichar c = scanner.peek();
            switch (c) {
            case '"':
            case '\'':
                title_delimiter = c;
                break;
            case '(':
                title_delimiter = ')';
                break;
            }

            if (title_delimiter != '\0') {
                state = State.TITLE;
                _title = new StringBuilder();
                scanner.next();
                if (!scanner.has_next()) {
                    _title.append("\n");
                }
            } else {
                finish_reference();
                /* There might be another reference instead, try that for the
                   same character. */
                state = State.START_DEFINITION;
            }
            return true;
        }

        private bool title(Scanner scanner) {
            var start = scanner.position();
            if (!LinkScanner
                .scan_link_title_content(scanner, title_delimiter))
                /* Invalid title, stop */
                return false;

            _title.append(scanner.get_source(start, scanner.position())
                          .get_content());

            if (!scanner.has_next()) {
                /* Title ran until the end of line, so continue on next line
                   (until we find the delimiter) */
                _title.append("\n");
                return true;
            }

            /* Skip delimiter character */
            scanner.next();
            scanner.whitespace();
            if (scanner.has_next()) {
                /* spec: No further non-whitespace characters may occur on the
                   line. */
                return false;
            }
            reference_valid = true;
            finish_reference();
            paragraph_lines.clear();

            /* See if there's another definition. */
            state = State.START_DEFINITION;
            return true;
        }

        private void finish_reference() {
            if (!reference_valid)
                return;

            var d = Escaping.unescape_string(_destination);
            var t = _title != null ?
                Escaping.unescape_string(_title.str) : null;
            var definition =
                new LinkReferenceDefinition.full(_label.str, d, t);
            definition.source_spans = source_spans;
            source_spans.clear();
            definitions.add(definition);

            _label = null;
            reference_valid = false;
            _destination = null;
            _title = null;
        }

        public enum State {
            /* Looking for the start of a definition, i.e. `[` */
            START_DEFINITION,
            /* Parsing the label, i.e. `foo` within `[foo]` */
            LABEL,
            /* Parsing the destination, i.e. `/url` in `[foo]: /url` */
            DESTINATION,
            /* Looking for the start of a title, i.e. the first `"` in `[foo]:
             * /url "title"` */
            START_TITLE,
            /* Parsing the content of the title, i.e. `title` in `[foo]: /url
             * "title"` */
            TITLE,

            /* End state, no matter what kind of lines we add, they won't be
             * references */
            PARAGRAPH;
        }
    }
}
