namespace Commonmark {

    public class LinkScanner : GLib.Object {

        /**
         * Attempt to scan the contents of a link label (inside the brackets),
         * stopping after the content or returning false.  The stopped position can
         * bei either the closing ], or the end of the line if the label continues
         * on the next line.
         */
        public static bool scan_link_label_content(Scanner scanner) {
            while (scanner.has_next()) {
                switch (scanner.peek()) {
                case '\\':
                    scanner.next();
                    if (Parsing.is_escapable(scanner.peek()))
                        scanner.next();
                    break;
                case ']':
                    return true;
                case '[':
                    /* spec: Unescaped square bracket characters are not allowed
                       inside the opening and closing square brackets of link
                       labels. */
                    return false;
                default:
                    scanner.next();
                    break;
                }
            }
            return true;
        }

        /**
         * Attempt to scan a link destination, stopping after the destination or
         * returning false.
         */
        public static bool scan_link_destination(Scanner scanner) {
            if (!scanner.has_next())
                return false;

            if (scanner.next_char('<')) {
                while (scanner.has_next()) {
                    switch (scanner.peek()) {
                    case '\\':
                        scanner.next();
                        if (Parsing.is_escapable(scanner.peek())) {
                            scanner.next();
                        }
                        break;
                    case '\n':
                    case '<':
                        return false;
                    case '>':
                        scanner.next();
                        return true;
                    default:
                        scanner.next();
                        break;
                    }
                }
                return false;
            } else {
                return scan_link_destination_with_balanced_parens(scanner);
            }
        }

        public static bool scan_link_title(Scanner scanner) {
            if (!scanner.has_next()) {
                return false;
            }

            unichar end_delimiter;
            switch (scanner.peek()) {
            case '"':
                end_delimiter = '"';
                break;
            case '\'':
                end_delimiter = '\'';
                break;
            case '(':
                end_delimiter = ')';
                break;
            default:
                return false;
            }
            scanner.next();

            if (!scan_link_title_content(scanner, end_delimiter)) {
                return false;
            }
            if (!scanner.has_next()) {
                return false;
            }
            scanner.next();
            return true;
        }

        public static bool scan_link_title_content(Scanner scanner,
                                                   unichar end_delimiter) {
            while (scanner.has_next()) {
                unichar c = scanner.peek();
                if (c == '\\') {
                    scanner.next();
                    if (Parsing.is_escapable(scanner.peek())) {
                        scanner.next();
                    }
                } else if (c == end_delimiter) {
                    return true;
                } else if (end_delimiter == ')' && c == '(') {
                    /* unescaped '(' in title within parens is invalid */
                    return false;
                } else {
                    scanner.next();
                }
            }
            return true;
        }

        /* spec: a nonempty sequence of characters that does not start with <, does
           not include ASCII space or control characters, and includes parentheses
           only if (a) they are backslash-escaped or (b) they are part of a balanced
           pair of unescaped parentheses */
        private static bool scan_link_destination_with_balanced_parens(
            Scanner scanner) {
            int parens = 0;
            bool empty = true;
            while (scanner.has_next()) {
                var c = scanner.peek();
                switch (c) {
                case ' ':
                    return !empty;
                case '\\':
                    scanner.next();
                    if (Parsing.is_escapable(scanner.peek())) {
                        scanner.next();
                    }
                    break;
                case '(':
                    parens++;
                    /* Limit to 32 nested parens for pathological cases */
                    if (parens > 32) {
                        return false;
                    }
                    scanner.next();
                    break;
                case ')':
                    if (parens == 0) {
                        return true;
                    } else {
                        parens--;
                    }
                    scanner.next();
                    break;
                default:
                    /* or control character */
                    /* FIXME */
                    /* if (Character.isISOControl(c)) {
                       return !empty;
                       } */
                    scanner.next();
                    break;
                }
                empty = false;
            }
            return true;
        }
    }
}
