namespace Commonmark {

    public class ListBlockParser : AbstractBlockParser {

        private ListBlock block;

        private bool had_blank_line;
        private int lines_after_blank;

        public ListBlockParser(ListBlock block) {
            this.block = block;
        }

        public override bool is_container() {
            return true;
        }

        public override bool can_contain(Block child_block) {
            if (child_block is ListItem) {
                /* Another list item is added to this list block. If the
                   previous line was blank, that means this list block is
                   "loose" (not tight).
                   spec: A list is loose if any of its constituent list items
                   are separated by blank lines */
                if (had_blank_line && lines_after_blank == 1) {
                    block.tight = false;
                    had_blank_line = false;
                }
                return true;
            } else {
                return false;
            }
        }

        public override Block get_block() {
            return block;
        }

        public override BlockContinue
            try_continue(ParserState state) {
            if (state.is_blank()) {
                had_blank_line = true;
                lines_after_blank = 0;
            } else if (had_blank_line) {
                lines_after_blank++;
            }
            /* List blocks themselves don't have any markers, only list
               items. So try to stay in the list.  If there is a block start
               other than list item, canContain makes sure that this list is
               closed. */
            return BlockContinue.at_index(state.get_index());
        }

        /**
         * Parse a list marker and return data on the marker or null.
         */
        private static ListData? parse_list(string line,
                                           int marker_index,
                                           int marker_column,
                                           bool in_paragraph) {
            var list_marker = parse_list_marker(line, marker_index);
            if (list_marker == null)
                return null;
            var list_block = list_marker.list_block;

            int index_after_marker = list_marker.index_after_marker;
            int marker_length = index_after_marker - marker_index;
            /* marker doesn't include tabs, so counting them as columns directly
               is ok */
            int column_after_marker = marker_column + marker_length;
            /* the column within the line where the content starts */
            int content_column = column_after_marker;

            /* See at which column the content starts if there is content */
            bool has_content = false;
            int length = line.length;
            for (int i = index_after_marker; i < length; i++) {
                char c = line.get(i);
                if (c == '\t') {
                    content_column +=
                        Parsing.columns_to_next_tab_stop(content_column);
                } else if (c == ' ') {
                    content_column++;
                } else {
                    has_content = true;
                    break;
                }
            }

            if (in_paragraph) {
                /* If the list item is ordered, the start number must be 1 to
                 * interrupt a paragraph. */
                if (list_block is OrderedList &&
                    ((OrderedList)list_block).start_number != 1)
                    return null;
                /* Empty list item can not interrupt a paragraph. */
                if (!has_content)
                    return null;
            }

            if (!has_content || (content_column - column_after_marker) >
                Parsing.CODE_BLOCK_INDENT) {
                /* If this line is blank or has a code block, default to 1 space
                   after marker */
                content_column = column_after_marker + 1;
            }

            return new ListData(list_block, content_column);
        }

        private static ListMarkerData?
            parse_list_marker(string line, int index) {
            unichar c = line.get(index);
            switch (c) {
                /* spec: A bullet list marker is a -, +, or * character. */
            case '-':
            case '+':
            case '*':
                if (is_space_tab_or_end(line, index + 1)) {
                    var bullet_list = new BulletList();
                    bullet_list.bullet_marker = c;
                    return new ListMarkerData(bullet_list, index + 1);
                } else {
                    return null;
                }
            default:
                return parse_ordered_list(line, index);
            }
        }

        /* spec: An ordered list marker is a sequence of 1–9 arabic digits
           (0-9), followed by either a `.` character or a `)` character. */
        private static ListMarkerData?
            parse_ordered_list(string line, int index) {
            int digits = 0;
            int length = line.length;
            for (int i = index; i < length; i++) {
                unichar c = line.get(i);
                switch (c) {
                case '0':
                case '1':
                case '2':
                case '3':
                case '4':
                case '5':
                case '6':
                case '7':
                case '8':
                case '9':
                    digits++;
                    if (digits > 9)
                        return null;
                    break;
                case '.':
                case ')':
                    if (digits >= 1 && is_space_tab_or_end(line, i + 1)) {
                        var number = line.slice(index, i);
                        var ordered_list = new OrderedList();
                        ordered_list.start_number = int.parse(number);
                        ordered_list.delimiter = c;
                        return new ListMarkerData(ordered_list, i + 1);
                    } else {
                        return null;
                    }
                default:
                    return null;
                }
            }
            return null;
        }

        private static bool is_space_tab_or_end(string line, int index) {
            if (index < line.length) {
                switch (line.get(index)) {
                case ' ':
                case '\t':
                    return true;
                default:
                    return false;
                }
            } else {
                return true;
            }
        }

        /**
         * Returns true if the two list items are of the same type,
         * with the same delimiter and bullet character. This is used
         * in agglomerating list items into lists.
         */
        private static bool lists_match(ListBlock a, ListBlock b) {
            if (a is BulletList && b is BulletList) {
                return (((BulletList)a).bullet_marker ==
                        ((BulletList)b).bullet_marker);
            } else if (a is OrderedList && b is OrderedList) {
                return (((OrderedList)a).delimiter ==
                        ((OrderedList)b).delimiter);
            }
            return false;
        }

        public class Factory : AbstractBlockParserFactory {

            public override BlockStart
                try_start(ParserState state,
                          MatchedBlockParser
                          matched_block_parser) {
                var matched = matched_block_parser.get_matched_block_parser();

                if (state.get_indent() >= Parsing.CODE_BLOCK_INDENT)
                    return BlockStart.none();
                int marker_index = state.get_next_non_space_index();
                int marker_column = state.get_column() + state.get_indent();
                bool in_paragraph = !matched_block_parser
                    .get_paragraph_lines().is_empty;
                var list_data =
                    parse_list(state.get_line().content, marker_index,
                               marker_column, in_paragraph);
                if (list_data == null)
                    return BlockStart.none();

                int new_column = list_data.content_column;
                var list_item_parser = new ListItemParser(new_column -
                                                          state.get_column());

                /* prepend the list block if needed */
                if (!(matched is ListBlockParser) ||
                    !lists_match((ListBlock)matched.get_block(),
                                  list_data.list_block)) {
                    var list_block_parser =
                        new ListBlockParser(list_data.list_block);
                    /* We start out with assuming a list is tight. If we find a
                       blank line, we set it to loose later. */
                    list_data.list_block.tight = true;

                    return BlockStart.from_two(list_block_parser,
                                                            list_item_parser)
                        .at_column(new_column);
                } else {
                    return BlockStart.from(list_item_parser)
                        .at_column(new_column);
                }
            }
        }

        private class ListData : GLib.Object {
            public ListBlock list_block;
            public int content_column;

            public ListData(ListBlock list_block,
                            int content_column) {
                this.list_block = list_block;
                this.content_column = content_column;
            }
        }

        private class ListMarkerData : GLib.Object {
            public ListBlock list_block;
            public int index_after_marker;

            public ListMarkerData(ListBlock list_block,
                                  int index_after_marker) {
                this.list_block = list_block;
                this.index_after_marker = index_after_marker;
            }
        }
    }
}
