namespace Commonmark {

    /**
     * The base class of all CommonMark AST nodes (Block and inlines).
     *
     * A node can have multiple children, and a parent (except for the root
     * node).
     */
    public abstract class Node : GLib.Object {

        public Node parent { get; set; }
        public Node first_child { get; private set; }
        public Node last_child { get; private set; }
        public Node prev { get; private set; }
        public Node next { get; private set; }
        public Gee.List<SourceSpan> source_spans { get; set; }

        public abstract void accept(Visitor visitor);

        public void append_child(Node child) {
            child.unlink();
            child.parent = this;
            if (last_child != null) {
                last_child.next = child;
                child.prev = last_child;
                last_child = child;
            } else {
                first_child = child;
                last_child = child;
            }
        }

        public void prepend_child(Node child) {
            child.unlink();
            child.parent = this;
            if (first_child != null) {
                first_child.prev = child;
                child.next = first_child;
                first_child = child;
            } else {
                first_child = child;
                last_child = child;
            }
        }

        public void unlink() {
            if (prev != null) {
                prev.next = next;
            } else if (parent != null) {
                parent.first_child = next;
            }
            if (next != null) {
                next.prev = prev;
            } else if (parent != null) {
                parent.last_child = prev;
            }
            parent = null;
            next = null;
            prev = null;
        }

        public void insert_after(Node sibling) {
            sibling.unlink();
            sibling.next = next;
            if (sibling.next != null) {
                sibling.next.prev = sibling;
            }
            sibling.prev = this;
            next = sibling;
            sibling.parent = parent;
            if (sibling.next == null) {
                sibling.parent.last_child = sibling;
            }
        }

        public void insert_before(Node sibling) {
            sibling.unlink();
            sibling.prev = prev;
            if (sibling.prev != null) {
                sibling.prev.next = sibling;
            }
            sibling.next = this;
            prev = sibling;
            sibling.parent = parent;
            if (sibling.prev == null) {
                sibling.parent.first_child = sibling;
            }
        }

        /**
         * Add a source span to the end of the list.
         * @param source_span the source span to add
         */
        public void add_source_span(SourceSpan source_span) {
            if (source_spans == null) {
                source_spans = new Gee.ArrayList<SourceSpan>();
            }
            this.source_spans.add(source_span);
        }

        public string to_string() {
            return get_class().get_name() + "{" + to_string_attributes() + "}";
        }

        protected virtual string to_string_attributes() {
            return "";
        }
    }
}
