namespace Commonmark {

    public class Parsing : GLib.Object {

        public const string TAG_NAME = "[A-Za-z][A-Za-z0-9-]*";
        public const string ATTRIBUTE_NAME = "[a-zA-Z_:][a-zA-Z0-9:._-]*";
        public const string UNQUOTED_VALUE = "[^\"'=<>`\\x00-\\x20]+";
        public const string SINGLE_QUOTED_VALUE = "'[^']*'";
        public const string DOUBLE_QUOTED_VALUE = "\"[^\"]*\"";
        public const string ATTRIBUTE_VALUE = "(?:" + UNQUOTED_VALUE + "|" +
            SINGLE_QUOTED_VALUE + "|" + DOUBLE_QUOTED_VALUE + ")";
        public const string ATTRIBUTE_VALUE_SPEC = "(?:" + "\\s*=" + "\\s*" +
            ATTRIBUTE_VALUE + ")";
        public const string ATTRIBUTE = "(?:" + "\\s+" + ATTRIBUTE_NAME +
            ATTRIBUTE_VALUE_SPEC + "?)";
        public const string OPEN_TAG = "<" + TAG_NAME + ATTRIBUTE + "*" + "\\s*/?>";
        public const string CLOSE_TAG = "</" + TAG_NAME + "\\s*[>]";
        public const int CODE_BLOCK_INDENT = 4;

        public static int columns_to_next_tab_stop(int column) {
            /* Tab stop is 4 */
            return 4 - (column % 4);
        }

        public static int find(char c, string s, int start_index) {
            int length = s.length;
            for (int i = start_index; i < length; i++) {
                if (s.get(i) == c) {
                    return i;
                }
            }
            return -1;
        }

        public static int find_line_break(string s, int start_index) {
            int length = s.length;
            for (int i = start_index; i < length; i++) {
                switch (s.get(i)) {
                case '\n':
                case '\r':
                    return i;
                }
            }
            return -1;
        }

        public static bool is_blank(string s) {
            return find_non_space(s, 0) == -1;
        }

        public static bool has_non_space(string s) {
            int length = s.length;
            int skipped = skip(' ', s, 0, length);
            return skipped != length;
        }

        public static bool is_letter(string s, int index) {
            return s.get_char(index).isalpha();
        }

        public static bool is_space_or_tab(string s, int index) {
            if (index < s.length) {
                switch (s.get(index)) {
                case ' ':
                case '\t':
                    return true;
                }
            }
            return false;
        }

        public static bool is_escapable(unichar c) {
            switch (c) {
            case '!':
            case '"':
            case '#':
            case '$':
            case '%':
            case '&':
            case '\'':
            case '(':
            case ')':
            case '*':
            case '+':
            case ',':
            case '-':
            case '.':
            case '/':
            case ':':
            case ';':
            case '<':
            case '=':
            case '>':
            case '?':
            case '@':
            case '[':
            case '\\':
            case ']':
            case '^':
            case '_':
            case '`':
            case '{':
            case '|':
            case '}':
            case '~':
                return true;
            }
            return false;
        }

        /* See https://spec.commonmark.org/0.29/#punctuation-character */
        public static bool is_punctuation_code_point(int code_point) {
            var c = (unichar)code_point;
            return c.ispunct();
        }

        public static bool is_whitespace_code_point(int code_point) {
            var c = (unichar)code_point;
            return c.isspace();
        }

        /**
         * Prepares the input line replacing ''\0''
         */
        public static string prepare_line(string line) {
            return line;
        }

        public static int skip(unichar skip, string s, int start_index,
                               int end_index) {
            for (int i = start_index; i < end_index; i++) {
                if (s.get(i) != skip) {
                    return i;
                }
            }
            return end_index;
        }

        public static int skip_backwards(char skip, string s, int start_index,
                                         int last_index) {
            for (int i = start_index; i >= last_index; i--) {
                if (s.get(i) != skip) {
                    return i;
                }
            }
            return last_index - 1;
        }

        public static int skip_space_tab(string s, int start_index,
                                         int end_index) {
            for (int i = start_index; i < end_index; i++) {
                switch (s.get(i)) {
                case ' ':
                case '\t':
                    break;
                default:
                    return i;
                }
            }
            return end_index;
        }

        public static int skip_space_tab_backwards(string s, int start_index,
                                                   int last_index) {
            for (int i = start_index; i >= last_index; i--) {
                switch (s.get(i)) {
                case ' ':
                case '\t':
                    break;
                default:
                    return i;
                }
            }
            return last_index - 1;
        }

        private static int find_non_space(string s, int start_index) {
            int length = s.length;
            for (int i = start_index; i < length; i++) {
                switch (s.get(i)) {
                case ' ':
                case '\t':
                case '\n':
                case '\x0b':
                case '\f':
                case '\r':
                    break;
                default:
                    return i;
                }
            }
            return -1;
        }
    }
}
