namespace Commonmark {

    public class Scanner {

        /**
         * Character representing the end of input source (or outside of the
         * text in case of the "previous" methods).
         *
         * Note that we can use NULL to represent this because CommonMark does
         * not allow those in the input (we replace them in the beginning of
         * parsing).
         */
        public const unichar END = '\0';

        /* Lines without newlines at the end. The scanner will yield `\n`
           between lines because they're significant for parsing and the final
           output. There is no `\n` after the last line.*/
        private Gee.List<SourceLine> lines;
        /* Which line we're at. */
        private int line_index;
        /* The index within the line. If index == length(), we pretend that
           there's a `\n` and only advance after we yield that. */
        private int index;

        /* Current line or "" if at the end of the lines (using "" instead of
           null saves a null check) */
        private SourceLine line;
        private int line_length = 0;

        public Scanner(Gee.List<SourceLine> lines, int line_index, int index) {
            this.line = SourceLine.from("", null);
            this.lines = lines;
            this.line_index = line_index;
            this.index = index;
            if (!lines.is_empty) {
                check_position(line_index, index);
                set_line(lines.get(line_index));
            }
        }

        public static Scanner from(SourceLines lines) {
            return new Scanner(lines.lines, 0, 0);
        }

        public unichar peek() {
            if (index < line_length) {
                return line.content.get(index);
            } else {
                if (line_index < lines.size - 1) {
                    return '\n';
                } else {
                    /* Don't return newline for end of last line */
                    return END;
                }
            }
        }

        public int peek_code_point() {
            if (index < line_length) {
                unichar c = line.content.get(index);
                if (is_high_surrogate(c) && index + 1 < line_length) {
                    unichar low = line.content.get(index + 1);
                    if (is_low_surrogate(low)) {
                        return to_code_point(c, low);
                    }
                }
                return (int)c;
            } else {
                if (line_index < lines.size - 1) {
                    return '\n';
                } else {
                    /* Don't return newline for end of last line */
                    return (int)END;
                }
            }
        }

        public int peek_previous_code_point() {
            if (index > 0) {
                int prev = index - 1;
                unichar c = line.content.get(prev);
                if (is_low_surrogate(c) && prev > 0) {
                    unichar high = line.content.get(prev - 1);
                    if (is_high_surrogate(high)) {
                        return to_code_point(high, c);
                    }
                }
                return (int)c;
            } else {
                if (line_index > 0) {
                    return '\n';
                } else {
                    return (int)END;
                }
            }
        }

        public bool has_next() {
            if (index < line_length) {
                return true;
            } else {
                /* No newline at end of last line */
                return line_index < lines.size - 1;
            }
        }

        public void next() {
            index++;
            if (index > line_length) {
                line_index++;
                if (line_index < lines.size) {
                    set_line(lines.get(line_index));
                } else {
                    set_line(SourceLine.from("", null));
                }
                index = 0;
            }
        }

        /**
         * Check if the specified unichar is next and advance the position.
         *
         * @param c the unichar to check (including newline characters)
         * @return true if matched and position was advanced, false otherwise
         */
        public bool next_char(unichar c) {
            if (peek() == c) {
                next();
                return true;
            } else {
                return false;
            }
        }

        /**
         * Check if we have the specified content on the line and advanced the
         * position. Note that if you want to match newline characters, use
         * #next(unichar).
         *
         * @param content the text content to match on a single line (excluding
         * newline characters)
         * @return true if matched and position was advanced, false otherwise
         */
        public bool next_content(string content) {
            if (index < line_length && index + content.length <= line_length) {
                /* Can't use startsWith because it's not available on string */
                for (int i = 0; i < content.length; i++) {
                    if (line.content.get(index + i) != content.get(i)) {
                        return false;
                    }
                }
                index += content.length;
                return true;
            } else {
                return false;
            }
        }

        public int match_multiple(unichar c) {
            int count = 0;
            while (peek() == c) {
                count++;
                next();
            }
            return count;
        }

        public int match(CharMatcher matcher) {
            int count = 0;
            while (matcher.matches(peek())) {
                count++;
                next();
            }
            return count;
        }

        public int whitespace() {
            int count = 0;
            while (true) {
                switch (peek()) {
                case ' ':
                case '\t':
                case '\n':
                case '\x0b':
                case '\f':
                case '\r':
                    count++;
                    next();
                    break;
                default:
                    return count;
                }
            }
        }

        public int find(unichar c) {
            int count = 0;
            while (true) {
                unichar cur = peek();
                if (cur == Scanner.END) {
                    return -1;
                } else if (cur == c) {
                    return count;
                }
                count++;
                next();
            }
        }

        public int find_matcher(CharMatcher matcher) {
            int count = 0;
            while (true) {
                unichar c = peek();
                if (c == END) {
                    return -1;
                } else if (matcher.matches(c)) {
                    return count;
                }
                count++;
                next();
            }
        }

        /* Don't expose the int index, because it would be good if we could
           switch input to a List<String> of lines later instead of one
           contiguous String. */
        public Position position() {
            return new Position(line_index, index);
        }

        public void set_position(Position position) {
            check_position(position.line_index, position.index);
            this.line_index = position.line_index;
            this.index = position.index;
            set_line(lines.get(this.line_index));
        }

        /* For cases where the caller appends the result to a StringBuilder, we
           could offer another method to avoid some unnecessary copying. */
        public SourceLines get_source(Position begin, Position end) {
            if (begin.line_index == end.line_index) {
                /* Shortcut for common case of text from a single line */
                SourceLine line = lines.get(begin.line_index);
                string new_content = line.content
                    .slice(begin.index, end.index);
                SourceSpan new_source_span = null;
                var source_span = line.source_span;
                if (source_span != null) {
                    new_source_span = SourceSpan.from(
                        source_span.line_index,
                        source_span.column_index + begin.index,
                        new_content.length);
                }
                return SourceLines.from_source_line(
                    SourceLine.from(new_content, new_source_span));
            } else {
                var source_lines = SourceLines.empty();
                var first_line = lines.get(begin.line_index);
                source_lines.add_line(
                    first_line.substring(
                        begin.index, first_line.content.length));

                /* Lines between begin and end (we are appending the full
                 * line) */
                for (int line = begin.line_index + 1; line < end.line_index; line++) {
                    source_lines.add_line(lines.get(line));
                }

                var last_line = lines.get(end.line_index);
                source_lines.add_line(last_line.substring(0, end.index));
                return source_lines;
            }
        }

        private void set_line(SourceLine line) {
            this.line = line;
            this.line_length = line.content.length;
        }

        private void check_position(int line_index, int index) {
            if (line_index < 0 || line_index >= lines.size)
                GLib.error("Line index %d out of range, number of lines: %d",
                           line_index, lines.size);
            var line = lines.get(line_index);
            if (index < 0 || index > line.content.length) {
                GLib.error("Index %d out of range, line length: %d",
                           index, line.content.length);
            }
        }

        private bool is_high_surrogate(unichar c) {
            return c >= 0xd800 && c <= 0xdbff;
        }

        private bool is_low_surrogate(unichar c) {
            return c >= 0xdc00 && c <= 0xdfff;
        }

        private int to_code_point(unichar high, unichar low) {
            return (int)(((high & 0xFFFF) << 16) & (low & 0xFFFF));
        }
    }
}
