namespace Commonmark {

    /**
     * A source span references a snippet of text from the source input.
     *
     * It has a starting position (line and column index) and a length of how
     * many characters it spans.
     *
     * For example, this CommonMark source text:
     * {{{
     *   > foo
     * }}}
     * The BlockQuote node would have this source span: line 0, column 0, length 5.
     *
     * The Paragraph node inside it would have: line 0, column 2, length 3.
     *
     * If a block has multiple lines, it will have a source span for each line.
     *
     * Note that the column index and length are measured in unichar characters
     * (UTF-16 code units). If you're outputting them to be consumed by another
     * programming language, e.g. one that uses UTF-8 strings, you will need to
     * translate them, otherwise characters such as emojis will result in
     * incorrect positions.
     */
    public class SourceSpan : GLib.Object {

        public int line_index { get; private set; }
        public int column_index { get; private set; }
        public int length { get; private set; }

        public static SourceSpan from(int line_index,
                                      int column_index,
                                      int length) {
            return new SourceSpan(line_index, column_index, length);
        }

        private SourceSpan(int line_index, int column_index, int length) {
            this.line_index = line_index;
            this.column_index = column_index;
            this.length = length;
        }

        public bool equals(GLib.Object o) {
            if (this == o)
                return true;
            if (o == null || this.get_type() != o.get_type())
                return false;
            var source_span = (SourceSpan)o;
            return line_index == source_span.line_index &&
                column_index == source_span.column_index &&
                length == source_span.length;
        }

        public int hash_code() {
            return line_index * column_index + length;
        }

        public string to_string() {
            return @"SourceSpan{line=$(line_index), column=$(column_index), " +
                "length=$(length)}";
        }
    }
}
