namespace Commonmark {

    /**
     * An implementation of DelimiterProcessor that dispatches all calls to two
     * or more other DelimiterProcessors depending on the length of the
     * delimiter run. All child DelimiterProcessors must have different minimum
     * lengths. A given delimiter run is dispatched to the child with the
     * largest acceptable minimum length. If no child is applicable, the one
     * with the largest minimum length is chosen.
     */
    public class StaggeredDelimiterProcessor : GLib.Object, DelimiterProcessor {

        private unichar delim;
        private int min_length;
        /* in reverse get_min_length order */
        private Gee.ArrayList<DelimiterProcessor> processors =
            new Gee.ArrayList<DelimiterProcessor>();

        public StaggeredDelimiterProcessor(unichar delim) {
            this.delim = delim;
        }

        public unichar get_opening_character() {
            return delim;
        }

        public unichar get_closing_character() {
            return delim;
        }

        public int get_min_length() {
            return min_length;
        }

        public void add(DelimiterProcessor dp) {
            int len = dp.get_min_length();
            var it = processors.bidir_list_iterator();
            var added = false;
            while (it.has_next()) {
                it.next();
                var p = it.get();
                int p_len = p.get_min_length();
                if (len > p_len) {
                    it.previous();
                    it.add(dp);
                    added = true;
                    break;
                } else if (len == p_len) {
                    var m = "Cannot add two delimiters processors for char " +
                        "'%d' and minimum length %d".printf((int)delim, len);
                    GLib.error(m);
                }
            }
            if (!added) {
                processors.add(dp);
                this.min_length = len;
            }
        }

        private DelimiterProcessor find_processor(int len) {
            foreach (var p in processors) {
                if (p.get_min_length() <= len)
                    return p;
            }
            return processors.get(0);
        }

        public int process(DelimiterRun opening_run, DelimiterRun closing_run) {
            return find_processor(opening_run.length()).process(opening_run,
                                                                closing_run);
        }
    }
}
