namespace Commonmark {

    public class TextContentRenderer : GLib.Object, Renderer {

        private bool _strip_newlines;

        public Gee.List<TextContentNodeRendererFactory> node_renderer_factories;

        private TextContentRenderer(Builder builder) {
            this._strip_newlines = builder._strip_newlines;

            this.node_renderer_factories =
                new Gee.ArrayList<TextContentNodeRendererFactory>();
            this.node_renderer_factories.add_all(builder.node_renderer_factories);
            /* Add as last. This means clients can override the rendering of
               core nodes if they want. */
            this.node_renderer_factories.add(new SimpleFactory());
        }

        /**
         * Create a new builder for configuring an {@link TextContentRenderer}.
         *
         * @return a builder
         */
        public static Builder builder() {
            return new Builder();
        }

        public void render_to(Node node, GLib.OutputStream output) {
            RendererContext context = new RendererContext(
                this, new TextContentWriter(output));
            context.render(node);
        }

        public string render(Node node) {
            if (node == null)
                GLib.error("node must not be null");
            try {
                var os = new GLib.MemoryOutputStream(null,
                                                     GLib.realloc,
                                                     GLib.free);
                render_to(node, os);
                os.close();
                var data = os.steal_data();
                data.length = (int)os.get_data_size();
                return (string)data;
            } catch (GLib.Error error) {
                GLib.error(error.message);
            }
        }

        /**
         * Builder for configuring an {@link TextContentRenderer}. See methods
         * for default configuration.
         */
        public class Builder : GLib.Object {

            public bool _strip_newlines = false;
            public Gee.List<TextContentNodeRendererFactory>
                node_renderer_factories =
                new Gee.ArrayList<TextContentNodeRendererFactory>();

            /**
             * @return the configured {@link TextContentRenderer}
             */
            public TextContentRenderer build() {
                return new TextContentRenderer(this);
            }

            /**
             * Set the value of flag for stripping new lines.
             *
             * @param _strip_newlines true for stripping new lines and render
             *                        text as "single line", false for keeping
             *                        all line breaks
             * @return ''this''
             */
            public Builder strip_newlines(bool _strip_newlines) {
                this._strip_newlines = _strip_newlines;
                return this;
            }

            /**
             * Add a factory for instantiating a node renderer (done when
             * rendering). This allows to override the rendering of node types
             * or define rendering for custom node types.
             *
             * If multiple node renderers for the same node type are created,
             * the one from the factory that was added first "wins". (This is
             * how the rendering for core node types can be overridden; the
             * default rendering comes last.)
             *
             * @param node_renderer_factory the factory for creating a node
             * renderer
             * @return ''this''
             */
            public Builder node_renderer_factory(
                TextContentNodeRendererFactory node_renderer_factory) {
                this.node_renderer_factories.add(node_renderer_factory);
                return this;
            }

            /**
             * @param extensions extensions to use on this text content renderer
             * @return ''this''
             */
            public Builder extensions(Gee.Iterable<Extension> extensions) {
                foreach (var extension in extensions) {
                    if (extension is TextContentRenderer.TextContentRendererExtension) {
                        var text_content_renderer_extension =
                            (TextContentRenderer.TextContentRendererExtension)extension;
                        text_content_renderer_extension.extend(this);
                    }
                }
                return this;
            }
        }

        /**
         * Extension for {@link TextContentRenderer}.
         */
        public interface TextContentRendererExtension :
            GLib.Object, Extension {
            public abstract void extend(
                TextContentRenderer.Builder renderer_builder);
        }

        public class RendererContext : GLib.Object,
            TextContentNodeRendererContext {

            private TextContentRenderer text_content_renderer;
            private TextContentWriter text_content_writer;
            private NodeRendererMap node_renderer_map = new NodeRendererMap();

            public RendererContext(TextContentRenderer text_content_renderer,
                                    TextContentWriter text_content_writer) {
                this.text_content_renderer = text_content_renderer;
                this.text_content_writer = text_content_writer;

                /* The first node renderer for a node type "wins". */
                for (int i = text_content_renderer.node_renderer_factories
                         .size - 1; i >= 0; i--) {
                    var node_renderer_factory =
                        text_content_renderer.node_renderer_factories.get(i);
                    var node_renderer = node_renderer_factory.create(this);
                    node_renderer_map.add(node_renderer);
                }
            }

            public bool strip_newlines() {
                return text_content_renderer._strip_newlines;
            }

            public TextContentWriter get_writer() {
                return text_content_writer;
            }

            public void render(Node node) {
                node_renderer_map.render(node);
            }
        }

        public class SimpleFactory : GLib.Object,
            TextContentNodeRendererFactory {

            public NodeRenderer create(TextContentNodeRendererContext context) {
                return new CoreTextContentNodeRenderer(context);
            }
        }
    }
}
