namespace Commonmark.Test {

    public class TestParser : Test.TestCase {

        public TestParser() {
            base("TestParser");
            add_test("test_custom_block_parser_factory",
                     test_custom_block_parser_factory);
            add_test("test_enabled_block_types",
                     test_enabled_block_types);
            add_test("test_indentantion", test_indentantion);
            add_test("test_inline_parser", test_inline_parser);
        }

        public void test_custom_block_parser_factory() {
            var parser = Parser.builder()
                .custom_block_parser_factory(new DashBlockParserFactory())
                .build();

            // The dashes would normally be a ThematicBreak
            var document = parser.parse("hey\n\n---\n");

            var first_child = document.first_child;
            GLib.assert_true(first_child is Paragraph);
            GLib.assert_cmpstr("hey",
                               GLib.CompareOperator.EQ,
                               ((Text)first_child.first_child).literal);
            var last_child = document.last_child;
            GLib.assert_true(last_child is DashBlock);
        }

        public void test_enabled_block_types() {
            var given = "# heading 1\n\nnot a heading";

            // all core parsers by default
            var parser = Parser.builder().build();
            var document = parser.parse(given);
            var first_child = document.first_child;
            GLib.assert_true(first_child is Heading);

            var headers_only = new Gee.HashSet<GLib.Type>();
            headers_only.add(typeof(Heading));
            parser = Parser.builder().set_enabled_block_types(headers_only)
                .build();
            document = parser.parse(given);
            GLib.assert_true(document.first_child is Heading);

            var no_core_types = new Gee.HashSet<GLib.Type>();
            parser = Parser.builder().set_enabled_block_types(no_core_types)
                .build();
            document = parser.parse(given);
            GLib.assert_false(document.first_child is Heading);
        }

        public void test_indentantion() {
            var given = (" - 1 space\n   - 3 spaces\n     " +
                         "- 5 spaces\n\t - tab + space");
            var parser = Parser.builder().build();
            var document = parser.parse(given);

            var list = document.first_child; /* first level list */
            GLib.assert_true(list is BulletList);

            GLib.assert_true(list.first_child == list.last_child);
            GLib.assert_cmpstr("1 space",
                               GLib.CompareOperator.EQ,
                               first_text(list.first_child));

            list = list.first_child.last_child; /* second level list */
            GLib.assert_true(list.first_child == list.last_child);
            GLib.assert_cmpstr("3 spaces",
                               GLib.CompareOperator.EQ,
                               first_text(list.first_child));

            list = list.first_child.last_child; /* third level list */
            GLib.assert_cmpstr("5 spaces",
                               GLib.CompareOperator.EQ,
                               first_text(list.first_child));
            GLib.assert_cmpstr("tab + space",
                               GLib.CompareOperator.EQ,
                               first_text(list.first_child.next));
        }

        public void test_inline_parser() {
            var fake_inline_parser_factory = new FakeInlineParserFactory();
            var parser = Parser.builder()
                .set_inline_parser_factory(fake_inline_parser_factory).build();
            var input = "**bold** **bold** ~~strikethrough~~";
            GLib.assert_true(parser.parse(input).first_child.first_child
                             is ThematicBreak);
        }

        private string first_text(Node n) {
            while (!(n is Text)) {
                GLib.assert_true(n != null);
                n = n.first_child;
            }
            return ((Text)n).literal;
        }

        private class FakeInlineParser : GLib.Object, InlineParser {

            public void parse(SourceLines lines, Node node) {
                node.append_child(new ThematicBreak());
            }
        }

        private class FakeInlineParserFactory : GLib.Object,
            InlineParserFactory {

            private FakeInlineParser fake_inline_parser;

            public FakeInlineParserFactory() {
                fake_inline_parser = new FakeInlineParser();
            }

            public InlineParser create(
                InlineParserContext inline_parser_context) {
                return fake_inline_parser;
            }
        }

        private class DashBlock : CustomBlock {}

        private class DashBlockParser : AbstractBlockParser {

            private DashBlock dash;

            construct {
                dash = new DashBlock();
            }

            public override Block get_block() {
                return dash;
            }

            public override BlockContinue
                try_continue(ParserState parser_state) {
                return BlockContinue.none();
            }
        }

        private class DashBlockParserFactory :
            AbstractBlockParserFactory {

            public override BlockStart
                try_start(ParserState state,
                          MatchedBlockParser matchedBlockParser) {
                if (state.get_line().content == "---") {
                    return BlockStart.from(
                        new DashBlockParser());
                }
                return BlockStart.none();
            }
        }
    }
}
